# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
from enum import Enum
from typing import Any, Generic, Optional, Tuple, Type, TypeVar

import attrs
import numpy as np

from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QSlider,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool import a121
from acconeer.exptool.a121._core.entities import Criticality


def widget_wrap_layout(layout: QLayout) -> QWidget:
    dummy = QWidget()
    dummy.setLayout(layout)
    return dummy


T = TypeVar("T")
EnumT = TypeVar("EnumT", bound=Enum)


@attrs.frozen(kw_only=True, slots=False)
class ParameterWidgetFactory(abc.ABC):
    name_label_text: str
    note_label_text: Optional[str] = None

    @abc.abstractmethod
    def create(self, parent: QWidget) -> ParameterWidget:
        ...


class ParameterWidget(QWidget):
    """Base class for a parameter-bound widget.

    A ``ParameterWidget`` comes with a
    ``name`` label and an ``note`` label by default.
    """

    sig_parameter_changed = QtCore.Signal(object)

    def __init__(self, factory: ParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(parent=parent)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._body_widget = QWidget(self)
        self.layout().addWidget(self._body_widget)

        self.__label_widget = QLabel(factory.name_label_text, parent=self._body_widget)

        self._body_layout = self._create_body_layout(self.__label_widget)
        self._body_widget.setLayout(self._body_layout)

        self.__note_widget = QLabel(parent=self)
        self.__note_widget.setWordWrap(True)
        self.__note_widget.setContentsMargins(5, 5, 5, 5)
        self.set_note_text(factory.note_label_text)
        self.layout().addWidget(self.__note_widget)

    def _create_body_layout(self, note_label_widget: QWidget) -> QLayout:
        """Called by ParameterWidget.__init__"""

        layout = QHBoxLayout(self._body_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(note_label_widget)
        return layout

    def set_note_text(
        self, message: Optional[str], criticality: Optional[Criticality] = None
    ) -> None:
        if not message:
            self.__note_widget.hide()
            return

        COLOR_MAP = {
            Criticality.ERROR: "#E6635A",
            Criticality.WARNING: "#FCC842",
            None: "white",
        }

        self.__note_widget.show()
        self.__note_widget.setText(message)
        self.__note_widget.setStyleSheet(
            f"background-color: {COLOR_MAP[criticality]}; color: white; font: bold italic;"
        )

    @abc.abstractmethod
    def set_parameter(self, value: Any) -> None:
        pass


@attrs.frozen(kw_only=True, slots=False)
class IntParameterWidgetFactory(ParameterWidgetFactory):
    limits: Optional[Tuple[Optional[int], Optional[int]]] = None
    suffix: Optional[str] = None

    def create(self, parent: QWidget) -> IntParameterWidget:
        return IntParameterWidget(self, parent)


class IntParameterWidget(ParameterWidget):
    def __init__(self, factory: IntParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self.__spin_box = _PidgetSpinBox(
            self._body_widget,
            limits=factory.limits,
            suffix=factory.suffix,
        )
        self.__spin_box.valueChanged.connect(self.__on_changed)
        self._body_layout.addWidget(self.__spin_box)

    def set_parameter(self, value: Any) -> None:
        assert isinstance(value, int)

        with QtCore.QSignalBlocker(self):
            self.__spin_box.setValue(value)

    def __on_changed(self) -> None:
        self.sig_parameter_changed.emit(self.__spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class FloatParameterWidgetFactory(ParameterWidgetFactory):
    limits: Optional[Tuple[Optional[float], Optional[float]]] = None
    suffix: Optional[str] = None
    decimals: int = 1

    def create(self, parent: QWidget) -> FloatParameterWidget:
        return FloatParameterWidget(self, parent)


class FloatParameterWidget(ParameterWidget):
    def __init__(self, factory: FloatParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self.__spin_box = _PidgetDoubleSpinBox(
            self._body_widget,
            limits=factory.limits,
            suffix=factory.suffix,
            decimals=factory.decimals,
        )
        self.__spin_box.valueChanged.connect(self.__on_changed)
        self._body_layout.addWidget(self.__spin_box)

    def set_parameter(self, value: Any) -> None:
        assert isinstance(value, (int, float))

        with QtCore.QSignalBlocker(self):
            self.__spin_box.setValue(value)

    def __on_changed(self) -> None:
        self.sig_parameter_changed.emit(self.__spin_box.value())


@attrs.frozen(kw_only=True, slots=False)
class FloatSliderParameterWidgetFactory(FloatParameterWidgetFactory):
    limits: Tuple[float, float]
    log_scale: bool = False
    show_limit_values: bool = True
    limit_texts: Optional[Tuple[Optional[str], Optional[str]]] = None

    def __attrs_post_init__(self) -> None:
        if self.log_scale:
            if self.limits[0] <= 0:
                raise ValueError("Lower limit must be > 0 when using log scale")

    def create(self, parent: QWidget) -> FloatSliderParameterWidget:
        return FloatSliderParameterWidget(self, parent)


class FloatSliderParameterWidget(ParameterWidget):
    def __init__(self, factory: FloatSliderParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self.__spin_box = _PidgetDoubleSpinBox(
            self._body_widget,
            limits=factory.limits,
            suffix=factory.suffix,
            decimals=factory.decimals,
        )
        self.__spin_box.valueChanged.connect(self.__on_spin_box_changed)
        self._body_layout.addWidget(self.__spin_box, 0, 1)

        slider_widget = QWidget(self._body_widget)
        self._body_layout.addWidget(slider_widget, 1, 0, 1, -1)
        slider_widget.setLayout(QHBoxLayout(slider_widget))
        slider_widget.layout().setContentsMargins(11, 6, 11, 0)

        lower_limit, upper_limit = factory.limits
        if factory.show_limit_values:
            slider_widget.layout().addWidget(QLabel(str(lower_limit), slider_widget))

        self.__slider = _PidgetFloatSlider(
            slider_widget,
            limits=factory.limits,
            decimals=factory.decimals,
            log_scale=factory.log_scale,
        )
        self.__slider.wrapped_value_changed.connect(self.__on_slider_changed)
        slider_widget.layout().addWidget(self.__slider)

        if factory.show_limit_values:
            slider_widget.layout().addWidget(QLabel(str(upper_limit), slider_widget))

        if factory.limit_texts is not None:
            label_text_widget = QWidget(self._body_widget)
            self._body_layout.addWidget(label_text_widget, 2, 0, 1, -1)
            label_text_widget.setLayout(QHBoxLayout(label_text_widget))
            label_text_widget.layout().setContentsMargins(11, 0, 11, 0)
            label_text_widget.layout().setSpacing(0)

            left, right = factory.limit_texts

            if left is not None:
                label = QLabel(left, label_text_widget)
                label_text_widget.layout().addWidget(label)

            label_text_widget.layout().addStretch(1)

            if right is not None:
                label = QLabel(right, label_text_widget)
                label_text_widget.layout().addWidget(label)

    def _create_body_layout(self, note_label_widget: QWidget) -> QLayout:
        """Called by ParameterWidget.__init__"""

        layout = QGridLayout(self._body_widget)
        layout.setContentsMargins(0, 0, 0, 6)
        layout.setSpacing(0)
        layout.addWidget(note_label_widget, 0, 0)
        return layout

    def set_parameter(self, value: Any) -> None:
        assert isinstance(value, (int, float))

        with QtCore.QSignalBlocker(self):
            self.__spin_box.setValue(value)
            self.__slider.wrapped_set_value(value)

    def __on_spin_box_changed(self, value: float) -> None:
        with QtCore.QSignalBlocker(self):
            self.__slider.wrapped_set_value(value)

        self.sig_parameter_changed.emit(value)

    def __on_slider_changed(self, value: float) -> None:
        with QtCore.QSignalBlocker(self):
            self.__spin_box.setValue(value)

        self.sig_parameter_changed.emit(value)


@attrs.frozen(kw_only=True, slots=False)
class OptionalParameterWidgetFactory(ParameterWidgetFactory):
    checkbox_label_text: Optional[str] = None


class OptionalParameterWidget(ParameterWidget):
    """Optional parameter, not optional widget"""

    def __init__(self, factory: OptionalParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._optional_widget = QWidget(self._body_widget)
        self._body_layout.addWidget(self._optional_widget)

        self._none_checkbox = QCheckBox(self._optional_widget)
        if factory.checkbox_label_text:
            self._none_checkbox.setText(factory.checkbox_label_text)

        self._optional_layout = self._create_optional_layout(self._none_checkbox)

    def _create_optional_layout(self, none_checkbox: QWidget) -> QLayout:
        """Called by OptionalParameterWidget.__init__"""

        layout = QHBoxLayout(self._optional_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addStretch(1)
        layout.addWidget(none_checkbox)
        return layout

    def set_parameter(self, value: Any) -> None:
        with QtCore.QSignalBlocker(self):
            self._none_checkbox.setChecked(value is not None)


@attrs.frozen(kw_only=True, slots=False)
class OptionalIntParameterWidgetFactory(OptionalParameterWidgetFactory, IntParameterWidgetFactory):
    init_set_value: Optional[int] = None

    def create(self, parent: QWidget) -> OptionalIntParameterWidget:
        return OptionalIntParameterWidget(self, parent)


class OptionalIntParameterWidget(OptionalParameterWidget):
    def __init__(self, factory: OptionalIntParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self.__spin_box = _PidgetSpinBox(
            self._optional_widget,
            limits=factory.limits,
            suffix=factory.suffix,
            init_set_value=factory.init_set_value,
        )
        self._optional_layout.addWidget(self.__spin_box)

        self._none_checkbox.stateChanged.connect(self.__on_changed)
        self.__spin_box.valueChanged.connect(self.__on_changed)

    def __on_changed(self) -> None:
        checked = self._none_checkbox.isChecked()

        with QtCore.QSignalBlocker(self):
            self.__spin_box.setEnabled(checked)

        value = self.__spin_box.value() if checked else None
        self.sig_parameter_changed.emit(value)

    def set_parameter(self, value: Any) -> None:
        super().set_parameter(value)

        with QtCore.QSignalBlocker(self):
            if value is None:
                self.__spin_box.setEnabled(False)
            else:
                self.__spin_box.setValue(value)
                self.__spin_box.setEnabled(True)


@attrs.frozen(kw_only=True, slots=False)
class OptionalFloatParameterWidgetFactory(
    OptionalParameterWidgetFactory, FloatParameterWidgetFactory
):
    init_set_value: Optional[float] = None

    def create(self, parent: QWidget) -> OptionalFloatParameterWidget:
        return OptionalFloatParameterWidget(self, parent)


class OptionalFloatParameterWidget(OptionalParameterWidget):
    def __init__(self, factory: OptionalFloatParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        self.__spin_box = _PidgetDoubleSpinBox(
            self._optional_widget,
            decimals=factory.decimals,
            limits=factory.limits,
            suffix=factory.suffix,
            init_set_value=factory.init_set_value,
        )
        self._optional_layout.addWidget(self.__spin_box)

        self._none_checkbox.stateChanged.connect(self.__on_changed)
        self.__spin_box.valueChanged.connect(self.__on_changed)

    def __on_changed(self) -> None:
        checked = self._none_checkbox.isChecked()

        with QtCore.QSignalBlocker(self):
            self.__spin_box.setEnabled(checked)

        value = self.__spin_box.value() if checked else None
        self.sig_parameter_changed.emit(value)

    def set_parameter(self, value: Any) -> None:
        super().set_parameter(value)

        with QtCore.QSignalBlocker(self):
            if value is None:
                self.__spin_box.setEnabled(False)
            else:
                self.__spin_box.setValue(value)
                self.__spin_box.setEnabled(True)


@attrs.frozen(kw_only=True, slots=False)
class CheckboxParameterWidgetFactory(ParameterWidgetFactory):
    def create(self, parent: QWidget) -> CheckboxParameterWidget:
        return CheckboxParameterWidget(self, parent)


class CheckboxParameterWidget(ParameterWidget):
    def __init__(self, factory: CheckboxParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)

        assert isinstance(self._body_layout, QGridLayout)

        self.__checkbox = QCheckBox(self._body_widget)
        self.__checkbox.clicked.connect(self.__on_checkbox_click)
        self._body_layout.addWidget(self.__checkbox, 0, 0)
        self._body_layout.setColumnStretch(0, 0)
        self._body_layout.setColumnStretch(1, 1)

    def _create_body_layout(self, note_label_widget: QWidget) -> QLayout:
        """Called by ParameterWidget.__init__"""

        layout = QGridLayout(self._body_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(note_label_widget, 0, 1)
        return layout

    def __on_checkbox_click(self, checked: bool) -> None:
        self.sig_parameter_changed.emit(checked)

    def set_parameter(self, param: Any) -> None:
        with QtCore.QSignalBlocker(self):
            self.__checkbox.setChecked(bool(param))


@attrs.frozen(kw_only=True, slots=False)
class ComboboxParameterWidgetFactory(ParameterWidgetFactory, Generic[T]):
    items: list[tuple[str, T]]

    def create(self, parent: QWidget) -> ComboboxParameterWidget[T]:
        return ComboboxParameterWidget(self, parent)


class ComboboxParameterWidget(ParameterWidget, Generic[T]):
    def __init__(self, factory: ComboboxParameterWidgetFactory[T], parent: QWidget) -> None:
        super().__init__(factory, parent)

        self._combobox = _PidgetComboBox(self._body_widget)
        self._body_layout.addWidget(self._combobox)

        for displayed_text, user_data in factory.items:
            self._combobox.addItem(displayed_text, user_data)

        self._combobox.currentIndexChanged.connect(self.__emit_data_of_combobox_item)

    def __emit_data_of_combobox_item(self, index: int) -> None:
        data = self._combobox.itemData(index)
        self.sig_parameter_changed.emit(data)

    def set_parameter(self, param: Any) -> None:
        with QtCore.QSignalBlocker(self):
            index = self._combobox.findData(param)
            if index == -1:
                raise ValueError(f"Data item {param} could not be found in {self}.")
            self._combobox.setCurrentIndex(index)


@attrs.frozen(kw_only=True, slots=False)
class UpdateableComboboxParameterWidgetFactory(ComboboxParameterWidgetFactory[T]):
    def create(self, parent: QWidget) -> UpdateableComboboxParameterWidget[T]:
        return UpdateableComboboxParameterWidget(self, parent)


class UpdateableComboboxParameterWidget(ComboboxParameterWidget[T]):
    def __init__(
        self, factory: UpdateableComboboxParameterWidgetFactory[T], parent: QWidget
    ) -> None:
        super().__init__(factory, parent)

    def update_items(self, items: list[tuple[str, T]]) -> None:
        with QtCore.QSignalBlocker(
            self
        ):  # Does not take into account when selected item is removed
            self._combobox.clear()

            for displayed_text, user_data in items:
                self._combobox.addItem(displayed_text, user_data)

    def set_parameter(self, param: Any) -> None:
        try:
            super().set_parameter(param)
        except ValueError:
            self.setEnabled(False)
        else:
            self.setEnabled(True)


@attrs.frozen(kw_only=True, slots=False)
class SensorIdParameterWidgetFactory(UpdateableComboboxParameterWidgetFactory[int]):
    name_label_text: str = attrs.field(default="Sensor:")

    def create(self, parent: QWidget) -> SensorIdParameterWidget:
        return SensorIdParameterWidget(self, parent)


class SensorIdParameterWidget(UpdateableComboboxParameterWidget[int]):
    _server_info: Optional[a121.ServerInfo]

    def __init__(self, factory: SensorIdParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)
        self._server_info = None

    def update_available_sensor_list(self, server_info: Optional[a121.ServerInfo]) -> None:
        if server_info == self._server_info:
            return

        self._server_info = server_info

        if server_info is None:
            self.update_items([])
            self.setEnabled(False)
        else:
            self.update_items(
                [
                    (str(sensor_id), sensor_id)
                    for sensor_id, sensor_info in server_info.sensor_infos.items()
                    if sensor_info.connected
                ]
            )
            self.setEnabled(True)


@attrs.frozen(kw_only=True, slots=False)
class EnumParameterWidgetFactory(ComboboxParameterWidgetFactory, Generic[EnumT]):
    enum_type: Type[EnumT] = attrs.field()
    label_mapping: dict[EnumT, str] = attrs.field()

    items: list[tuple[str, EnumT]] = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        if self.label_mapping.keys() != set(self.enum_type):
            raise ValueError("label_mapping does not match enum_type")

        items = [(v, k) for k, v in self.label_mapping.items()]

        # The instance is immutable at this point, which is circumvented by the next row. See:
        # - https://www.attrs.org/en/stable/api.html#attr.ib
        # - https://github.com/python-attrs/attrs/issues/120
        # - https://github.com/python-attrs/attrs/issues/147

        object.__setattr__(self, "items", items)

    def create(self, parent: QWidget) -> EnumParameterWidget[EnumT]:
        return EnumParameterWidget[EnumT](self, parent)


class EnumParameterWidget(ComboboxParameterWidget[EnumT]):
    def __init__(self, factory: EnumParameterWidgetFactory, parent: QWidget) -> None:
        super().__init__(factory, parent)


_WIDGET_WIDTH = 125


class _PidgetComboBox(QComboBox):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setFixedWidth(_WIDGET_WIDTH)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class _PidgetSpinBox(QSpinBox):
    def __init__(
        self,
        parent: QWidget,
        *,
        limits: Optional[Tuple[Optional[int], Optional[int]]] = None,
        init_set_value: Optional[int] = None,
        suffix: Optional[str] = None,
    ) -> None:
        super().__init__(parent)

        self.setKeyboardTracking(False)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setAlignment(QtCore.Qt.AlignRight)
        self.setFixedWidth(_WIDGET_WIDTH)

        self.setRange(*_convert_int_limits_to_qt_range(limits))

        if suffix:
            self.setSuffix(f" {suffix}")

        if init_set_value is not None:
            self.setValue(init_set_value)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class _PidgetDoubleSpinBox(QDoubleSpinBox):
    def __init__(
        self,
        parent: QWidget,
        *,
        limits: Optional[Tuple[Optional[float], Optional[float]]] = None,
        init_set_value: Optional[float] = None,
        decimals: int = 1,
        suffix: Optional[str] = None,
    ) -> None:
        super().__init__(parent)

        self.setKeyboardTracking(False)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setAlignment(QtCore.Qt.AlignRight)
        self.setFixedWidth(_WIDGET_WIDTH)

        self.setRange(*_convert_float_limits_to_qt_range(limits))
        self.setDecimals(decimals)
        self.setSingleStep(10 ** (-decimals))

        if suffix:
            self.setSuffix(f" {suffix}")

        if init_set_value is not None:
            self.setValue(init_set_value)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()


class _PidgetFloatSlider(QSlider):
    NUM_STEPS = 1000

    wrapped_value_changed = QtCore.Signal(float)

    def __init__(
        self,
        parent: QWidget,
        *,
        limits: Tuple[float, float],
        decimals: int,
        log_scale: bool,
    ) -> None:
        super().__init__(QtCore.Qt.Horizontal, parent)

        self.limits = limits
        self.decimals = decimals
        self.log_scale = log_scale

        self.setRange(0, self.NUM_STEPS)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

        self.valueChanged.connect(self.__on_value_changed)

    def __on_value_changed(self, value: int) -> None:
        self.wrapped_value_changed.emit(self.__from_slider_scale(value))

    def wrapped_set_value(self, value: float) -> None:
        self.setValue(self.__to_slider_scale(value))

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if self.hasFocus():
            super().wheelEvent(event)
        else:
            event.ignore()

    def __to_slider_scale(self, x: float) -> int:
        lower, upper = self.limits

        if self.log_scale:
            lower = np.log(lower)
            upper = np.log(upper)
            x = np.log(x)

        y = (x - lower) / (upper - lower) * self.NUM_STEPS
        return int(round(y))

    def __from_slider_scale(self, y: int) -> float:
        lower, upper = self.limits

        if self.log_scale:
            lower = np.log(lower)
            upper = np.log(upper)

        x = y / self.NUM_STEPS * (upper - lower) + lower

        if self.log_scale:
            x = np.exp(x)

        return round(x, self.decimals)


def _convert_int_limits_to_qt_range(
    limits: Optional[Tuple[Optional[int], Optional[int]]]
) -> Tuple[int, int]:
    if limits is None:
        limits = (None, None)

    lower, upper = limits

    if lower is None:
        lower = int(-1e9)

    if upper is None:
        upper = int(1e9)

    return (lower, upper)


def _convert_float_limits_to_qt_range(
    limits: Optional[Tuple[Optional[float], Optional[float]]]
) -> Tuple[float, float]:
    if limits is None:
        limits = (None, None)

    lower, upper = limits

    if lower is None:
        lower = -1e9

    if upper is None:
        upper = 1e9

    return (lower, upper)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional

from PySide6.QtWidgets import QGridLayout, QGroupBox, QHBoxLayout, QVBoxLayout, QWidget


class VerticalGroupBox(QGroupBox):
    def __init__(self, title: str, parent: Optional[QWidget]) -> None:
        super().__init__(parent=parent)
        self.setTitle(title)
        self.setLayout(QVBoxLayout(parent=self))


class HorizontalGroupBox(QGroupBox):
    def __init__(self, title: str, parent: Optional[QWidget]) -> None:
        super().__init__(parent=parent)
        self.setTitle(title)
        self.setLayout(QHBoxLayout(parent=self))


class GridGroupBox(QGroupBox):
    def __init__(self, title: str, parent: Optional[QWidget]) -> None:
        super().__init__(parent=parent)
        self.setTitle(title)
        self.setLayout(QGridLayout(parent=self))



from ctypes import *
from ctypes.wintypes import *


_ole32 = oledll.ole32

_StringFromCLSID = _ole32.StringFromCLSID
_CoTaskMemFree = windll.ole32.CoTaskMemFree


"""Flags controlling what is included in the device information set built by SetupDiGetClassDevs"""
DIGCF_DEFAULT = 0x00000001
DIGCF_PRESENT = 0x00000002
DIGCF_ALLCLASSES = 0x00000004
DIGCF_PROFILE = 0x00000008
DIGCF_DEVICE_INTERFACE = 0x00000010

"""Flags controlling File acccess"""
GENERIC_WRITE = 1073741824
GENERIC_READ = -2147483648
FILE_SHARE_READ = 1
FILE_SHARE_WRITE = 2
OPEN_EXISTING = 3
OPEN_ALWAYS = 4
FILE_ATTRIBUTE_NORMAL = 128
FILE_FLAG_OVERLAPPED = 1073741824

INVALID_HANDLE_VALUE = HANDLE(-1)

""" USB PIPE TYPE """
PIPE_TYPE_CONTROL = 0
PIPE_TYPE_ISO = 1
PIPE_TYPE_BULK = 2
PIPE_TYPE_INTERRUPT = 3


""" Errors """
ERROR_IO_INCOMPLETE = 996
ERROR_IO_PENDING = 997


class UsbSetupPacket(Structure):
    _fields_ = [
        ("request_type", c_ubyte),
        ("request", c_ubyte),
        ("value", c_ushort),
        ("index", c_ushort),
        ("length", c_ushort),
    ]


class Overlapped(Structure):
    _fields_ = [
        ("Internal", LPVOID),
        ("InternalHigh", LPVOID),
        ("Offset", DWORD),
        ("OffsetHigh", DWORD),
        ("Pointer", LPVOID),
        ("hEvent", HANDLE),
    ]


class UsbInterfaceDescriptor(Structure):
    _fields_ = [
        ("b_length", c_ubyte),
        ("b_descriptor_type", c_ubyte),
        ("b_interface_number", c_ubyte),
        ("b_alternate_setting", c_ubyte),
        ("b_num_endpoints", c_ubyte),
        ("b_interface_class", c_ubyte),
        ("b_interface_sub_class", c_ubyte),
        ("b_interface_protocol", c_ubyte),
        ("i_interface", c_ubyte),
    ]


class PipeInfo(Structure):
    _fields_ = [
        (
            "pipe_type",
            c_ulong,
        ),
        ("pipe_id", c_ubyte),
        ("maximum_packet_size", c_ushort),
        ("interval", c_ubyte),
    ]


class LpSecurityAttributes(Structure):
    _fields_ = [
        ("n_length", DWORD),
        ("lp_security_descriptor", c_void_p),
        ("b_Inherit_handle", BOOL),
    ]


class GUID(Structure):
    _fields_ = [("data1", DWORD), ("data2", WORD), ("data3", WORD), ("data4", c_byte * 8)]

    def __repr__(self):
        return 'GUID("%s")' % str(self)

    def __str__(self):
        p = c_wchar_p()
        _StringFromCLSID(byref(self), byref(p))
        result = p.value
        _CoTaskMemFree(p)
        return result

    def __cmp__(self, other):
        if isinstance(other, GUID):
            a = bytes(self)
            b = bytes(other)
            return (a > b) - (a < b)
        return -1

    def __nonzero__(self):
        return self != GUID_null

    def __eq__(self, other):
        return isinstance(other, GUID) and bytes(self) == bytes(other)

    def __hash__(self):
        # We make GUID instances hashable, although they are mutable.
        return hash(bytes(self))


GUID_null = GUID()


class SpDevinfoData(Structure):
    _fields_ = [
        ("cb_size", DWORD),
        ("class_guid", GUID),
        ("dev_inst", DWORD),
        ("reserved", POINTER(c_ulong)),
    ]


class SpDeviceInterfaceData(Structure):
    _fields_ = [
        ("cb_size", DWORD),
        ("interface_class_guid", GUID),
        ("flags", DWORD),
        ("reserved", POINTER(c_ulong)),
    ]


class SpDeviceInterfaceDetailData(Structure):
    _fields_ = [("cb_size", DWORD), ("device_path", WCHAR * 1)]  # devicePath array!!!



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
import logging
from typing import Generic, Optional, TypeVar

import pyqtgraph as pg

from acconeer.exptool.app.new import (
    AppModel,
    BackendPlugin,
    GeneralMessage,
    PlotPlugin,
    ViewPlugin,
)


log = logging.getLogger(__name__)


T = TypeVar("T")


class A121BackendPluginBase(Generic[T], BackendPlugin[T]):
    pass


class A121ViewPluginBase(ViewPlugin):
    pass


class A121PlotPluginBase(PlotPlugin):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)
        self._is_setup = False
        self._plot_job: Optional[GeneralMessage] = None

    def handle_message(self, message: GeneralMessage) -> None:
        if message.name == "setup":
            self.plot_layout.clear()
            self.setup_from_message(message)
            self._is_setup = True
        elif message.name == "plot":
            self._plot_job = message
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.update_from_message(self._plot_job)
        finally:
            self._plot_job = None

    @abc.abstractmethod
    def setup_from_message(self, message: GeneralMessage) -> None:
        pass

    @abc.abstractmethod
    def update_from_message(self, message: GeneralMessage) -> None:
        pass



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Type

import numpy as np

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorPlotPluginBase,
    ProcessorViewPluginBase,
)
from acconeer.exptool.a121.algo.virtual_button import (
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_sensor_config,
)
from acconeer.exptool.app.new import AppModel, Plugin, PluginFamily, PluginGeneration
from acconeer.exptool.app.new.ui.plugin import PidgetFactoryMapping, pidgets


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):
    @classmethod
    def get_processor_cls(cls) -> Type[Processor]:
        return Processor

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        # Note: Incomplete mapping
        return {
            "sensitivity_close": pidgets.FloatParameterWidgetFactory(
                name_label_text="Sensitivity",
                decimals=1,
                limits=(0.1, 4),
            ),
            "patience_close": pidgets.IntParameterWidgetFactory(
                name_label_text="Patience",
                limits=(0, None),
            ),
            "calibration_duration_s": pidgets.FloatParameterWidgetFactory(
                name_label_text="Calibration duration",
                suffix="s",
                limits=(0, None),
            ),
            "calibration_interval_s": pidgets.FloatParameterWidgetFactory(
                name_label_text="Calibration interval",
                suffix="s",
                limits=(1, None),
            ),
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig


class PlotPlugin(ProcessorPlotPluginBase[ProcessorResult]):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)

    def setup(self, metadata: a121.Metadata, sensor_config: a121.SensorConfig) -> None:
        self.detection_history_plot = self._create_detection_plot(self.plot_layout)

        self.detection_history_curve_close = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(1, width=5)
        )
        self.detection_history_curve_far = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(0, width=5)
        )

        close_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Close detection")
        )
        far_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Far detection")
        )
        self.close_text_item = pg.TextItem(
            html=close_html,
            fill=pg.mkColor(0xFF, 0x7F, 0x0E),
            anchor=(0.5, 0),
        )
        self.far_text_item = pg.TextItem(
            html=far_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4),
            anchor=(0.5, 0),
        )
        pos_left = (100 / 3, 1.8)
        pos_right = (2 * 100 / 3, 1.8)
        self.close_text_item.setPos(*pos_left)
        self.far_text_item.setPos(*pos_right)
        self.detection_history_plot.addItem(self.close_text_item)
        self.detection_history_plot.addItem(self.far_text_item)
        self.close_text_item.hide()
        self.far_text_item.hide()

        self.detection_history = np.full((2, 100), np.NaN)

    def update(self, processor_result: ProcessorResult) -> None:
        detection = np.array([processor_result.detection_close, processor_result.detection_far])
        self.detection_history = np.roll(self.detection_history, -1, axis=1)
        self.detection_history[:, -1] = detection

        self.detection_history_curve_close.setData(self.detection_history[0])
        self.detection_history_curve_far.setData(self.detection_history[1])

        if processor_result.detection_close:
            self.close_text_item.show()
        else:
            self.close_text_item.hide()

        if processor_result.detection_far:
            self.far_text_item.show()
        else:
            self.far_text_item.hide()

    @staticmethod
    def _create_detection_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        detection_history_plot = parent.addPlot()
        detection_history_plot.setMenuEnabled(False)
        detection_history_plot.setMouseEnabled(x=False, y=False)
        detection_history_plot.hideButtons()
        detection_history_plot.showGrid(x=True, y=True, alpha=0.5)
        detection_history_plot.setYRange(-0.1, 1.8)
        return detection_history_plot


VIRTUAL_BUTTON_PLUGIN = Plugin(
    generation=PluginGeneration.A121,
    key="virtual_button",
    title="Virtual button",
    description=None,
    family=PluginFamily.EXAMPLE_APP,
    backend_plugin=BackendPlugin,
    plot_plugin=PlotPlugin,
    view_plugin=ViewPlugin,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
import logging
from typing import Dict, Generic, List

import pyqtgraph as pg

from acconeer.exptool import a121
from acconeer.exptool.a121.algo._base import MetadataT, ResultT
from acconeer.exptool.a121.algo._plugins._a121 import A121PlotPluginBase
from acconeer.exptool.app.new import AppModel, GeneralMessage


log = logging.getLogger(__name__)


class GenericProcessorPlotPluginBase(Generic[ResultT, MetadataT], A121PlotPluginBase):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)

    def setup_from_message(self, message: GeneralMessage) -> None:
        assert message.kwargs is not None
        self.setup(**message.kwargs)

    def update_from_message(self, message: GeneralMessage) -> None:
        self.update(message.data)  # type: ignore[arg-type]

    @abc.abstractmethod
    def setup(self, metadata: MetadataT, sensor_config: a121.SensorConfig) -> None:
        pass

    @abc.abstractmethod
    def update(self, processor_result: ResultT) -> None:
        pass


ProcessorPlotPluginBase = GenericProcessorPlotPluginBase[ResultT, a121.Metadata]
ExtendedProcessorPlotPluginBase = GenericProcessorPlotPluginBase[
    ResultT, List[Dict[int, a121.Metadata]]
]



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import attrs
import numpy as np
import numpy.typing as npt

from .common import attrs_ndarray_eq
from .result import Result, ResultContext
from .utils import get_subsweeps_from_frame, int16_complex_array_to_complex


NDArrayBool = npt.NDArray[np.bool_]
NDArrayInt = npt.NDArray[np.int64]


@attrs.frozen(kw_only=True)
class StackedResults:
    """Stacked results

    For loading/processing data.

    Representation of multiple :class:`Result` stacked together. Scalar values, like
    :attr:`Result.data_saturated`, become 1-D arrays of the same type. The :attr:`Result.frame`,
    which is originally a 2-D array, becomes a 3-D array where the frames are stacked in the first
    dimension.

    See :class:`Result` for details on the attributes/properties.
    """

    data_saturated: NDArrayBool = attrs.field(eq=attrs_ndarray_eq)
    frame_delayed: NDArrayBool = attrs.field(eq=attrs_ndarray_eq)
    calibration_needed: NDArrayBool = attrs.field(eq=attrs_ndarray_eq)
    temperature: NDArrayBool = attrs.field(eq=attrs_ndarray_eq)
    _frame: npt.NDArray = attrs.field(eq=attrs_ndarray_eq)

    tick: NDArrayInt = attrs.field()

    _context: ResultContext = attrs.field()

    @property
    def frame(self) -> npt.NDArray[np.complex_]:
        return int16_complex_array_to_complex(self._frame)

    @property
    def subframes(self) -> list[npt.NDArray[np.complex_]]:
        return get_subsweeps_from_frame(self.frame, self._context.metadata)

    @property
    def tick_time(self) -> npt.NDArray[np.float_]:
        return self.tick / self._context.ticks_per_second

    def __len__(self) -> int:
        return len(self._frame)

    def __getitem__(self, key: int) -> Result:
        return Result(
            calibration_needed=self.calibration_needed[key],
            data_saturated=self.data_saturated[key],
            frame_delayed=self.frame_delayed[key],
            temperature=self.temperature[key],
            tick=self.tick[key],
            frame=self._frame[key],
            context=self._context,
        )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from ._enums import (
    ConnectionInterface,
    ConnectionState,
    PluginFamily,
    PluginGeneration,
    PluginState,
)
from ._exceptions import HandledException
from .app import main
from .app_model import AppModel, PlotPlugin, Plugin, ViewPlugin
from .backend import BackendPlugin, GeneralMessage, Message, PluginStateMessage, is_task
from .storage import get_temp_dir, get_temp_h5_path
from .ui import BUTTON_ICON_COLOR



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from ._bootloader_tool import BootloaderTool
from ._meta import (
    ACCONEER_XC120_BOARD_PROTOCOL_PID,
    ACCONEER_XC120_BOOTLOADER_PID,
    ACCONEER_XC120_EXPLORATION_SERVER_PID,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from acconeer.exptool import a121


# Here we create a H5Recorder.
# The H5Recorder is an object that saves frames directly to a H5-file.
filename = "data.h5"
h5_recorder = a121.H5Recorder(filename)

# Client creation
client = a121.Client(ip_address="192.168.0.1")
client.connect()

# Session setup, just like the other examples.
config = a121.SessionConfig()
client.setup_session(config)

# Here we send the H5Recorder to the Client.
# After this call, the client is responsible for the H5Recorder.
# The Client will automatically close the file when
# "stop_session" is called.
client.start_session(recorder=h5_recorder)

n = 10
for i in range(n):
    # Client will send its Results to the H5Recorder.
    client.get_next()
    print(f"Result {i + 1}/{n} was sampled")

client.stop_session()
client.disconnect()

with a121.open_record(filename) as record:
    print(record)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from pathlib import Path
from typing import Optional

import qtawesome as qta

from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import QFileDialog, QHBoxLayout, QPushButton, QWidget

from acconeer.exptool.app.new.app_model import AppModel

from .misc import BUTTON_ICON_COLOR


class RecordingWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.layout().addWidget(LoadFileButton(app_model, self))
        self.layout().addWidget(SaveFileButton(app_model, self))


class FileButton(QPushButton):
    def __init__(
        self,
        app_model: AppModel,
        text: str,
        icon: QtGui.QIcon,
        shortcut: Optional[str],
        parent: QWidget,
    ) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.setText(text)
        self.setIcon(icon)

        if shortcut is not None:
            self.setShortcut(shortcut)

        self.clicked.connect(self._on_click)

        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)


class LoadFileButton(FileButton):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(
            app_model,
            "Load from file",
            qta.icon("fa.folder-open", color=BUTTON_ICON_COLOR),
            "ctrl+o",
            parent,
        )
        app_model.sig_notify.connect(self._on_app_model_update)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        self.setEnabled(app_model.plugin_state.is_steady)

    def _on_click(self) -> None:
        filename, _ = QFileDialog.getOpenFileName(
            self,
            caption="Load from file",
            filter=_file_suffix_to_filter(".h5", ".npz"),
            options=QFileDialog.DontUseNativeDialog,
        )

        if not filename:
            return

        self.app_model.load_from_file(Path(filename))


class SaveFileButton(FileButton):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(
            app_model,
            "Save to file",
            qta.icon("mdi.content-save", color=BUTTON_ICON_COLOR),
            "ctrl+s",
            parent,
        )
        app_model.sig_notify.connect(self._on_app_model_update)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        self.setEnabled(app_model.saveable_file is not None and app_model.plugin_state.is_steady)

    def _on_click(self) -> None:
        if self.app_model.saveable_file is None:
            raise RuntimeError

        filename, _ = QFileDialog.getSaveFileName(
            self,
            caption="Save to file",
            filter=_file_suffix_to_filter(self.app_model.saveable_file.suffix),
            options=QFileDialog.DontUseNativeDialog,
        )

        if not filename:
            return

        path = Path(filename)

        if path.suffix != self.app_model.saveable_file.suffix:
            path = path.with_name(path.name + self.app_model.saveable_file.suffix)

        self.app_model.save_to_file(path)


def _file_suffix_to_filter(*exts: str) -> str:
    NAMES = {
        ".h5": "HDF5",
        ".npz": "NumPy",
    }
    return ";;".join([f"{NAMES.get(ext, '')} (*{ext})" for ext in exts])



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from typing_extensions import Protocol


class Link(Protocol):
    def connect(self) -> None:
        """Establishes a connection."""
        ...

    @property
    def timeout(self) -> float:
        """Return link timout."""
        ...

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        """Set return link timeout."""
        ...

    def recv(self, num_bytes: int) -> bytes:
        """Recieves `num_bytes` bytes."""
        ...

    def send(self, bytes_: bytes) -> None:
        """Sends all `bytes_` over the link."""
        ...

    def disconnect(self) -> None:
        """Tears down the connection."""
        ...


class BufferedLink(Link, Protocol):
    def recv_until(self, byte_sequence: bytes) -> bytes:
        """Collects all bytes until `byte_sequence` is encountered,
        returning what was collected
        """
        ...



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Tuple

from typing_extensions import Protocol

from acconeer.exptool.a121._core.entities import (
    Metadata,
    Result,
    SensorInfo,
    ServerInfo,
    SessionConfig,
)


class CommunicationProtocol(Protocol):
    end_sequence: bytes

    @classmethod
    def get_system_info_command(cls) -> bytes:
        """The `get_system_info` command."""
        ...

    @classmethod
    def get_system_info_response(
        cls, bytes_: bytes, sensor_infos: dict[int, SensorInfo]
    ) -> Tuple[ServerInfo, str]:
        """Reads the response of `get_system_info` and parses it to a `ServerInfo`."""
        ...

    @classmethod
    def get_sensor_info_command(cls) -> bytes:
        """The `get_sensor_info` command."""
        ...

    @classmethod
    def get_sensor_info_response(cls, bytes_: bytes) -> dict[int, SensorInfo]:
        """Reads the response of `get_sensor_info` and returns
        a dict of the mapping sensor_id -> SensorInfo
        """
        ...

    @classmethod
    def setup_command(cls, session_config: SessionConfig) -> bytes:
        """The `setup` command."""
        ...

    @classmethod
    def setup_response(
        cls, bytes_: bytes, context_session_config: SessionConfig
    ) -> list[dict[int, Metadata]]:
        """Parses the reponse of the `setup` command and parses it to an extended Metadata."""
        ...

    @classmethod
    def start_streaming_command(cls) -> bytes:
        """The `start_streaming` command."""
        ...

    @classmethod
    def start_streaming_response(cls, bytes_: bytes) -> None:
        """Response of `start_streaming` command. May raise an exception if
        the server could not start streaming
        """
        ...

    @classmethod
    def get_next_header(
        cls, bytes_: bytes, extended_metadata: list[dict[int, Metadata]], ticks_per_second: int
    ) -> Tuple[int, list[dict[int, Result]]]:
        """Parses the header of a data package. Returns the payload size and
        partial (incomplete) Results.
        """
        ...

    @classmethod
    def get_next_payload(
        cls, bytes_: bytes, partial_results: list[dict[int, Result]]
    ) -> list[dict[int, Result]]:
        """Parses the payload of the data package. Populates the partial (incomplete)
        Results with data from the payload.
        """
        ...

    @classmethod
    def stop_streaming_command(cls) -> bytes:
        """The `stop_streaming` command"""
        ...

    @classmethod
    def stop_streaming_response(cls, bytes_: bytes) -> None:
        """Response of `stop_streaming` command. May raise an exception if
        the server could not stop streaming
        """
        ...



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

from PySide6 import QtGui

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.processing_config = processing_config

        self.downsampling_factor = sensor_config.downsampling_factor
        self.sweeps_per_frame = sensor_config.sweeps_per_frame
        sweep_rate = session_info["sweep_rate"]
        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.step_length = session_info["step_length_m"]
        self.f_res = sweep_rate / self.sweeps_per_frame
        self.fft_x_scale = 100 * self.step_length

        self.smooth_max = et.utils.SmoothMax(
            sweep_rate / self.sweeps_per_frame,
            tau_grow=0,
            tau_decay=0.5,
            hysteresis=0.1,
        )

        self.setup_is_done = False

    def setup(self, win):
        self.plots = []
        self.curves = []

        self.layout = win.layout
        self.layout.setRowStretchFactor(0, 2)
        self.layout.setRowStretchFactor(1, 3)
        self.layout.setVerticalSpacing(50)

        for i in range(self.depths.size):
            title = "{:.0f} cm".format(100 * self.depths[i])
            plot = win.addPlot(row=0, col=i, title=title)
            plot.setMenuEnabled(False)
            plot.setMouseEnabled(x=False, y=False)
            plot.hideButtons()
            plot.setYRange(0, 2**16)
            plot.hideAxis("left")

            if self.depths.size < 8:
                plot.getAxis("bottom").setStyle(stopAxisAtTick=(True, True))
                plot.getAxis("bottom").setTicks(
                    [
                        [
                            (0, "1"),
                            (self.sweeps_per_frame - 1, str(self.sweeps_per_frame)),
                        ],
                    ]
                )
                plot.setLabel("bottom", "Sweep Index")
            else:
                plot.hideAxis("bottom")

            plot.plot(np.arange(self.sweeps_per_frame), 2**15 * np.ones(self.sweeps_per_frame))
            curve = plot.plot(pen=et.utils.pg_pen_cycler())
            self.plots.append(plot)
            self.curves.append(curve)

        self.ft_plot = win.addPlot(row=1, col=0, colspan=self.depths.size)
        self.ft_plot.setMenuEnabled(False)
        self.ft_plot.setMouseEnabled(x=False, y=False)
        self.ft_plot.hideButtons()
        self.ft_im = pg.ImageItem(autoDownsample=True)
        self.ft_im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
        self.ft_plot.addItem(self.ft_im)
        self.ft_plot.setLabel("bottom", "Depth (cm)")
        self.ft_plot.getAxis("bottom").setTickSpacing(6 * self.downsampling_factor, 6)

        self.setup_is_done = True
        self.update_processing_config()

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        for plot in self.plots:
            plot.setVisible(self.processing_config.show_data_plot)

        if self.processing_config.show_data_plot is False:
            self.layout.setRowStretchFactor(0, 0)
            self.layout.setRowStretchFactor(1, 1)
            self.layout.setVerticalSpacing(0)

        else:
            self.layout.setRowStretchFactor(0, 2)
            self.layout.setRowStretchFactor(1, 3)
            self.layout.setVerticalSpacing(50)

        half_wavelength = 2.445e-3
        self.ft_im.resetTransform()
        tr = QtGui.QTransform()
        tr.translate(100 * (self.depths[0] - self.step_length / 2), 0)
        if self.processing_config.show_speed_plot:
            self.ft_plot.setLabel("left", "Speed (m/s)")
            tr.scale(self.fft_x_scale, self.f_res * half_wavelength)
        else:
            self.ft_plot.setLabel("left", "Frequency (kHz)")
            tr.scale(self.fft_x_scale, self.f_res * 1e-3)

        self.ft_im.setTransform(tr)

    def update(self, data):
        frame = data["frame"]

        for i, ys in enumerate(frame.T):
            self.curves[i].setData(ys)

        m = np.max(data["abs_fft"])
        m = max(m, 1e4)
        m = self.smooth_max.update(m)
        self.ft_im.updateImage(data["abs_fft"], levels=(0, m * 1.05))



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import argparse
import platform

import serial

import acconeer.exptool as et
from acconeer.exptool.flash._products import PRODUCT_FLASH_MAP


def flash_image(image_path, flash_port):
    if flash_port:
        if isinstance(flash_port, et.utils.USBDevice):
            flash_pid = flash_port.pid
            flash_device = flash_port
            flash_product = flash_port.name
        else:
            flash_pid = flash_port.pid
            flash_device = flash_port.device
            flash_product = flash_port.product

        if flash_pid in PRODUCT_FLASH_MAP.keys():
            PRODUCT_FLASH_MAP[flash_pid].flash(flash_device, image_path)
        else:
            raise NotImplementedError(f"No flash support for Acconeer device {flash_product}")
    else:
        raise ValueError("Flash port is None")


def find_flash_port(port=None):
    flash_port = None
    detected_ports = [
        pinfo
        for pinfo in et.utils.tag_serial_ports_objects(serial.tools.list_ports.comports())
        if pinfo[1]
    ]

    if len(detected_ports) == 0 and platform.system().lower() == "windows":
        detected_ports = et.utils.get_usb_devices()

    if len(detected_ports) == 0:
        print("No devices connected")
    elif port:
        for pinfo in detected_ports:
            if isinstance(pinfo, et.utils.USBDevice):
                if port == pinfo:
                    flash_port = pinfo
            elif port == pinfo[0].device:
                flash_port = pinfo[0]
        if not flash_port:
            print(f"{port} in not connected")
    elif len(detected_ports) == 1:
        if isinstance(detected_ports[0], et.utils.USBDevice):
            flash_port = detected_ports[0]
        else:
            flash_port = detected_ports[0][0]
    elif len(detected_ports) > 1:
        print("Found multiple Acconeer products:", end="\n\n")
        if isinstance(detected_ports[0], et.utils.USBDevice):
            for port in detected_ports:
                print(f"\t{port.name}")
        else:
            detected_ports = [(p.device, t) for (p, t) in detected_ports]
            for port, tag in [("Serial port:", "Model:")] + detected_ports:
                print(f"\t{port:<15} {tag:}")
        print('\nRun the script again and specify port using the "--port" flag.')

    return flash_port


def main():
    parser = argparse.ArgumentParser(description="Image Flasher")
    parser.add_argument(
        "--port", dest="port", metavar="port", help="Serial port or USB device name"
    )
    parser.add_argument("--image", "-i", required=True, help="Image file to flash")

    verbosity_group = parser.add_mutually_exclusive_group(required=False)
    verbosity_group.add_argument(
        "-v",
        "--verbose",
        action="store_true",
    )
    verbosity_group.add_argument(
        "-vv",
        "--debug",
        action="store_true",
    )
    verbosity_group.add_argument(
        "-q",
        "--quiet",
        action="store_true",
    )

    args = parser.parse_args()

    et.utils.config_logging(args)

    flash_port = find_flash_port(args.port)

    flash_image(args.image, flash_port)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from PySide6.QtWidgets import QWidget

from acconeer.exptool.app.new.app_model import AppModel


class AppModelAwareWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        app_model.sig_notify.connect(self.on_app_model_update)

    def on_app_model_update(self, app_model: AppModel) -> None:
        pass



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from argparse import ArgumentParser

from acconeer.exptool.setup.base import PlatformInstall


class SetupArgumentParser(ArgumentParser):
    def __init__(self) -> None:
        super().__init__(
            prog="python -m acconeer.exptool.setup",
            description="Helps you setup your computer for use with Acconeer products.",
        )

        self.add_argument("--platform", choices=PlatformInstall.platforms(), default=None)
        self.add_argument(
            "--silent",
            action="store_true",
            default=False,
            help="Skips printing information and yes/no prompts",
        )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
from typing import Any, Iterator

import numpy as np
import numpy.typing as npt

from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121._core.entities.configs import SessionConfig

from .client_info import ClientInfo
from .metadata import Metadata
from .result import Result
from .server_info import ServerInfo
from .stacked_results import StackedResults


class Record(abc.ABC):
    """Record representing all data needed to recreate a session"""

    @property
    @abc.abstractmethod
    def client_info(self) -> ClientInfo:
        """Client info"""

    @property
    @abc.abstractmethod
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        """Extended metadata"""

    @property
    def metadata(self) -> Metadata:
        """Retrieves the sole metadata entry in the record

        :raises: ValueError if there are multiple entries in the :attr:`extended_metadata`
        """

        return utils.unextend(self.extended_metadata)

    @property
    @abc.abstractmethod
    def extended_results(self) -> Iterator[list[dict[int, Result]]]:
        """The sequence of extended results

        Retrieves an iterator over all extended results in the record, which may be useful to
        replay a session.

        .. tip::

            If your record only has one entry in the session, and you want to retrieve the results
            for that one entry, you can use :attr:`results`.

            If you desire to get, for example, all frames in the sequence of results stacked in one
            array, use
            :attr:`extended_stacked_results`. Similar to :attr:`results`, there is also
            :attr:`stacked_results` which retrieves the one and only entry.
        """

    @property
    def results(self) -> Iterator[Result]:
        """Retrieves the sole sequence of results in the record

        :raises: ValueError if there are multiple entries in the :attr:`extended_results`
        """

        for extended_result in self.extended_results:
            yield utils.unextend(extended_result)

    @property
    @abc.abstractmethod
    def extended_stacked_results(self) -> list[dict[int, StackedResults]]:
        """The extended stacked results"""

    @property
    def stacked_results(self) -> StackedResults:
        """Retrieves the sole stacked results in the record

        :raises: ValueError if there are multiple entries in the :attr:`extended_stacked_results`
        """

        return utils.unextend(self.extended_stacked_results)

    @property
    def frames(self) -> npt.NDArray[np.complex_]:
        """Retrieves the sole stack of frames in the record

        Alias for ``Record.stacked_results.frame``.

        :raises: ValueError if there are multiple entries in the :attr:`extended_stacked_results`
        """

        return self.stacked_results.frame

    @property
    @abc.abstractmethod
    def lib_version(self) -> str:
        """The version of the ``acconeer.exptool`` library which created the record"""

    @property
    @abc.abstractmethod
    def num_frames(self) -> int:
        """The number of frames in the record"""

    @property
    @abc.abstractmethod
    def server_info(self) -> ServerInfo:
        """Server info"""

    @property
    @abc.abstractmethod
    def session_config(self) -> SessionConfig:
        """Session config"""

    @property
    @abc.abstractmethod
    def sensor_id(self) -> int:
        """The sole sensor ID

        :raises: ValueError if there are multiple entries in the :attr:`session_config`
        """

    @property
    @abc.abstractmethod
    def timestamp(self) -> str:
        """Creation timestamp"""

    @property
    @abc.abstractmethod
    def uuid(self) -> str:
        """UUID"""


class PersistentRecord(Record):
    """Record that wraps a file on disk

    Data is lazily loaded from the underlying file on demand.
    """

    @abc.abstractmethod
    def close(self) -> None:
        """Close the underlying file"""

    def __enter__(self) -> Record:
        return self

    def __exit__(self, *_: Any) -> None:
        self.close()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import AlgoConfigBase, AlgoParamEnum, ProcessorBase


class AmplitudeMethod(AlgoParamEnum):
    COHERENT = "Coherent"
    NONCOHERENT = "Noncoherent"
    FFT_MAX = "FFT max"


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoConfigBase):
    amplitude_method: AmplitudeMethod = attrs.field(
        default=AmplitudeMethod.COHERENT, converter=AmplitudeMethod
    )


@attrs.frozen(kw_only=True)
class ProcessorResult:
    frame: npt.NDArray[np.complex_]
    distance_velocity_map: npt.NDArray[np.float_]
    amplitudes: npt.NDArray[np.float_]
    phases: npt.NDArray[np.float_]


class Processor(ProcessorBase[ProcessorConfig, ProcessorResult]):
    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
    ) -> None:
        self.processor_config = processor_config
        spf = sensor_config.sweeps_per_frame
        self.window = np.hanning(spf)[:, None]
        self.window /= np.sum(self.window)

    def process(self, result: a121.Result) -> ProcessorResult:
        frame = result.frame

        z_ft = np.fft.fftshift(np.fft.fft(frame * self.window, axis=0), axes=(0,))
        abs_z_ft = np.abs(z_ft)

        amplitude_method = self.processor_config.amplitude_method
        if amplitude_method == AmplitudeMethod.COHERENT:
            ampls = np.abs(frame.mean(axis=0))
        elif amplitude_method == AmplitudeMethod.NONCOHERENT:
            ampls = np.abs(frame).mean(axis=0)
        elif amplitude_method == AmplitudeMethod.FFT_MAX:
            ampls = abs_z_ft.mean(axis=0)
        else:
            raise RuntimeError(f"Unknown AmplitudeMethod: {amplitude_method}")

        phases = np.angle(frame.mean(axis=0))

        return ProcessorResult(
            frame=frame,
            distance_velocity_map=abs_z_ft,
            amplitudes=ampls,
            phases=phases,
        )

    def update_config(self, config: ProcessorConfig) -> None:
        self.processor_config = config


def get_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        sweeps_per_frame=32,
        num_points=40,
        step_length=8,
    )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

"""Internal tool for updating the register map YAML file

Requirements: PyYAML and oyaml

Quick install:
python -m pip install --user pyyaml oyaml

Usage:
python update_regmap.py /path/to/sw-main/module_server/register_map/register_map.yaml
"""

import argparse
import importlib.util
import os

import oyaml as yaml


here = os.path.dirname(os.path.realpath(__file__))
exptool_package_path = os.path.join(here, "..", "src", "acconeer", "exptool")

modes_module_path = os.path.join(exptool_package_path, "modes.py")
modes_module_spec = importlib.util.spec_from_file_location(
    "acconeer.exptool.modes", modes_module_path
)
modes_module = importlib.util.module_from_spec(modes_module_spec)
modes_module_spec.loader.exec_module(modes_module)

parser = argparse.ArgumentParser()
parser.add_argument("input_filename")
args = parser.parse_args()
in_fn = args.input_filename

out_fn = os.path.join(exptool_package_path, "data", "regmap.yaml")
assert os.path.exists(os.path.dirname(out_fn))

with open(in_fn, "r") as in_f:
    d = yaml.full_load(in_f)


def clean(d):
    to_pop = []

    for k, v in d.items():
        if k == "description":
            to_pop.append(k)
            continue

        if type(v) == dict:
            if v.get("internal", False):
                to_pop.append(k)
                continue

            if "modes" in v:
                modes = v["modes"]
                assert isinstance(modes, (list, str))
                if isinstance(modes, str):
                    modes = [modes]

                for mode in modes[:]:
                    try:
                        modes_module.get_mode(mode)
                    except ValueError:
                        modes.remove(mode)

                if len(modes) == 0:
                    to_pop.append(k)
                    continue

                v["modes"] = modes

            clean(v)

    for k in to_pop:
        d.pop(k)


clean(d)

with open(out_fn, "w") as out_f:
    yaml.dump(d, out_f, default_flow_style=False)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import enum
import json
import logging
import warnings
from copy import deepcopy
from time import time

import numpy as np

from acconeer.exptool.a111._clients import links
from acconeer.exptool.a111._clients.base import (
    BaseClient,
    ClientError,
    SessionSetupError,
    decode_version_str,
)
from acconeer.exptool.a111._modes import Mode, get_mode


log = logging.getLogger(__name__)


class JsonProtocolBase:
    def __init__(self, link):
        self._link = link
        self._num_sensors = None
        self._sweeps_per_frame = None
        self._session_cmd = None
        self._mode = None

    def _send_cmd(self, cmd_dict):
        cmd_dict["api_version"] = 3
        s = json.dumps(cmd_dict, separators=(",", ":"))
        packed = bytearray(s + "\n", "ascii")

        self._link.send(packed)

    def _recv_frame(self):
        packed = self._link.recv_until(b"\n")
        header = json.loads(str(packed, "ascii"))
        payload_len = header["payload_size"]

        if payload_len > 0:
            payload = self._link.recv(payload_len)
        else:
            payload = None

        return header, payload

    def get_version(self):
        cmd = {"cmd": "get_version"}
        self._send_cmd(cmd)

        try:
            header, _ = self._recv_frame()
        except links.LinkError as e:
            raise ClientError("no response from server") from e

        log.debug("connected and got a response")

        msg = None
        if header["status"] == "ok":
            msg = header["message"].lower()
            log.info("version msg: {}".format(msg))
        else:
            log.info("No get_version cmd, try get_system_info")

        return msg

    def get_sensor_count(self):
        self._send_cmd({"cmd": "get_board_sensor_count"})
        header, _ = self._recv_frame()
        return int(header["message"])

    def setup_session(self, config):
        pass

    def init_session(self, retry=True):
        pass

    def start_session(self):
        if not self._session_ready:
            self.init_session()

        cmd = {"cmd": "start_streaming"}
        self._send_cmd(cmd)
        header, _ = self._recv_frame()
        if header["status"] != "start":
            raise ClientError

        log.debug("started streaming")

    def get_next(self):
        header, payload = self._recv_frame()

        status = header["status"]
        if status == "end":
            raise ClientError("session ended")
        elif status != "ok":
            raise ClientError("server error")

        info = self.decode_stream_header(header)
        data = self.decode_stream_payload(payload)
        return info, data

    def stop_session(self):
        pass

    def decode_stream_header(self, header):
        pass

    def decode_stream_payload(self, payload):
        pass


class JsonProtocolStreamingServer(JsonProtocolBase):
    def __init__(self, link, squeeze):
        super().__init__(link)
        self._squeeze = squeeze
        # stacklevel=5 will warn user code as deprecated.
        # (JsonPSS -> SocketClient -> ClientFactory -> Client -> <user code>)
        warnings.warn(
            "Streaming Server is deprecated. Consider upgrading to Exploration Server.",
            DeprecationWarning,
            stacklevel=5,
        )

    def setup_session(self, config):
        if isinstance(config, dict):
            cmd = deepcopy(config)
            log.warning("setup with raw dict config - you're on your own")
        else:
            cmd = self._get_dict_for_config(config)

        try:
            self._mode = get_mode(cmd["cmd"].lower().replace("_data", ""))
        except ValueError:
            self._mode = None

        self._sweeps_per_frame = cmd.get("sweeps_per_frame", 16)
        self._num_sensors = len(cmd["sensors"])

        cmd["output_format"] = "json+binary"
        self._session_cmd = cmd

        return cmd

    def init_session(self, retry=True):
        if self._session_cmd is None:
            raise ClientError

        self._send_cmd(self._session_cmd)
        header, _ = self._recv_frame()

        if header["status"] == "error":
            if retry:
                return self.init_session(retry=False)
            else:
                raise SessionSetupError
        elif header["status"] != "ok":
            raise ClientError("got unexpected header")

        log.debug("session initialized")

        self._session_ready = True
        info = self._get_session_info_for_header(header)

        return info

    def stop_session(self):
        cmd = {"cmd": "stop_streaming"}
        self._send_cmd(cmd)

        t0 = time()
        while time() - t0 < self._link.timeout:
            header, _ = self._recv_frame()
            status = header["status"]
            if status == "end":
                break
            elif status == "ok":  # got streaming data
                continue
            else:
                raise ClientError
        else:
            raise ClientError

        self._link.timeout = self._link.DEFAULT_TIMEOUT

        self._session_ready = False

        log.debug("stopped streaming")

    def decode_stream_header(self, header):
        raw_infos = header["result_info"]
        mapped_infos = [{} for _ in raw_infos]

        for (raw_info, mapped_info) in zip(raw_infos, mapped_infos):
            for raw_key, val in raw_info.items():
                mapped_key = DATA_HEADER_TO_INFO_KEY_REMAP.get(raw_key, raw_key)

                if mapped_key is None:
                    continue

                mapped_info[mapped_key] = val

        if self.squeeze and len(mapped_infos) == 1:
            return mapped_infos[0]
        else:
            return mapped_infos

    def decode_stream_payload(self, payload):
        if not payload:
            return None

        squeeze = self.squeeze and self._num_sensors == 1

        if self._mode == Mode.SPARSE:
            data = np.frombuffer(payload, dtype=">u2").astype("float")

            if squeeze:
                shape = (self._sweeps_per_frame, -1)
            else:
                shape = (self._num_sensors, self._sweeps_per_frame, -1)

            return data.reshape(shape)

        if self._mode == Mode.IQ:
            data = np.frombuffer(payload, dtype=">i2").astype("float")
            data = data.reshape((-1, 2)).view(dtype="complex").flatten()
        elif self._mode in (Mode.ENVELOPE, Mode.POWER_BINS):
            data = np.frombuffer(payload, dtype=">u2").astype("float")
        else:  # Fallback
            data = np.frombuffer(payload, dtype=">u2")

        if not squeeze:
            data = data.reshape((self._num_sensors, -1))

        return data

    def _get_dict_for_config(self, config):
        d = {}
        d["cmd"] = get_mode(config.mode).value.lower() + "_data"

        for config_key, cmd_key in CONFIG_TO_CMD_KEY_MAP.items():
            config_val = getattr(config, config_key, None)

            if self._mode == Mode.IQ and config_key == "sampling_mode":
                continue

            if config_val is None:
                continue

            if isinstance(config_val, bool):
                cmd_val = int(config_val)
            elif isinstance(config_val, enum.Enum):
                if hasattr(config_val, "json_value"):
                    cmd_val = config_val.json_value
                else:
                    cmd_val = config_val.value
            else:
                cmd_val = config_val

            d[cmd_key] = cmd_val

        return d

    def _get_session_info_for_header(self, header):
        info = {}

        for header_key, val in header.items():
            info_key = SESSION_HEADER_TO_INFO_KEY_REMAP.get(header_key, header_key)

            if info_key is None:
                continue

            info[info_key] = val

        return info

    @property
    def squeeze(self):
        return self._squeeze

    @squeeze.setter
    def squeeze(self, squeeze):
        self._squeeze = squeeze


class JsonProtocolExplorationServer(JsonProtocolBase):
    def __init__(self, link, squeeze):
        super().__init__(link)
        self._squeeze = squeeze

    def get_system_info(self):
        self._send_cmd({"cmd": "get_system_info"})
        try:
            header, _ = self._recv_frame()
        except links.LinkError as e:
            raise ClientError("no response from server") from e

        if header["status"] != "ok":
            raise ClientError(f"system_info error {header}")

        system_info = header["system_info"]
        return system_info

    def _set_baudrate(self, baudrate):
        set_baudrate_cmd = {"cmd": "set_uart_baudrate", "baudrate": baudrate}
        self._send_cmd(set_baudrate_cmd)
        try:
            header, _ = self._recv_frame()
        except links.LinkError as e:
            raise ClientError("no response from server") from e

        if header["status"] == "ok":
            self._link.baudrate = baudrate

    def setup_session(self, config):
        if isinstance(config, dict):
            cmd = deepcopy(config)
            log.warning("setup with raw dict config - you're on your own")
            return cmd

        cmd = {}
        cmd["cmd"] = "setup"
        service = get_mode(config.mode).value.lower()
        self._mode = get_mode(config.mode)

        update_rate = getattr(config, "update_rate", None)
        if update_rate:
            cmd["update_rate"] = update_rate

        repetition_mode = getattr(config, "repetition_mode", None)
        if repetition_mode:
            cmd["repetition_mode"] = repetition_mode.json_value

        cmd["groups"] = []
        cmd["groups"].append([])

        sensors = getattr(config, "sensor", None)
        sensor_config = {}

        for config_key, cmd_key in CONFIG_TO_CMD_KEY_MAP.items():
            config_val = getattr(config, config_key, None)

            if config_key in ["sensor", "update_rate", "repetition_mode"]:
                continue

            if self._mode == Mode.IQ and config_key == "sampling_mode":
                continue

            if config_val is None:
                continue

            if isinstance(config_val, bool):
                cmd_val = int(config_val)
            elif isinstance(config_val, enum.Enum):
                if hasattr(config_val, "json_value"):
                    cmd_val = config_val.json_value
                else:
                    cmd_val = config_val.value
            else:
                cmd_val = config_val

            sensor_config[cmd_key] = cmd_val

        for sensor in sensors:
            sensor_setup = {}
            sensor_setup["sensor_id"] = sensor
            sensor_setup["service"] = service
            sensor_setup["config"] = sensor_config
            cmd["groups"][0].append(sensor_setup)

        self._sweeps_per_frame = cmd["groups"][0][0]["config"].get("sweeps_per_frame", 16)
        self._num_sensors = len(sensors)

        self._session_cmd = cmd

        return cmd

    def init_session(self, retry=True):
        if self._session_cmd is None:
            raise ClientError

        self._send_cmd(self._session_cmd)
        header, _ = self._recv_frame()

        if header["status"] == "error":
            if retry:
                return self.init_session(retry=False)
            else:
                raise SessionSetupError
        elif header["status"] != "ok":
            raise ClientError("got unexpected header")

        log.debug("session initialized")

        self._session_ready = True

        info = {}

        for header_key, val in header["metadata"][0][0].items():
            info_key = SESSION_HEADER_TO_INFO_KEY_REMAP.get(header_key, header_key)

            if info_key is None:
                continue

            info[info_key] = val

        return info

    def stop_session(self):
        cmd = {"cmd": "stop_streaming"}
        self._send_cmd(cmd)

        t0 = time()
        while time() - t0 < self._link.timeout:
            header, _ = self._recv_frame()
            status = header["status"]
            if status == "stop":
                break
            elif status == "ok":  # got streaming data
                continue
            else:
                raise ClientError
        else:
            raise ClientError

        self._link.timeout = self._link.DEFAULT_TIMEOUT

        self._session_ready = False

        log.debug("stopped streaming")

    def decode_stream_header(self, header):
        raw_infos = header["result_info"][0]
        mapped_infos = [{} for _ in raw_infos]

        for (raw_info, mapped_info) in zip(raw_infos, mapped_infos):
            for raw_key, val in raw_info.items():
                mapped_key = DATA_HEADER_TO_INFO_KEY_REMAP.get(raw_key, raw_key)

                if mapped_key is None:
                    continue

                mapped_info[mapped_key] = val

        if self.squeeze and len(mapped_infos) == 1:
            return mapped_infos[0]
        else:
            return mapped_infos

    def decode_stream_payload(self, payload):
        if not payload:
            return None

        squeeze = self.squeeze and self._num_sensors == 1

        if self._mode == Mode.SPARSE:
            data = np.frombuffer(payload, dtype="<u2").astype("float")

            if squeeze:
                shape = (self._sweeps_per_frame, -1)
            else:
                shape = (self._num_sensors, self._sweeps_per_frame, -1)

            return data.reshape(shape)

        if self._mode == Mode.IQ:
            data = np.frombuffer(payload, dtype="<i2").astype("float")
            data = data.reshape((-1, 2)).view(dtype="complex").flatten()
        elif self._mode in (Mode.ENVELOPE, Mode.POWER_BINS):
            data = np.frombuffer(payload, dtype="<u2").astype("float")
        else:  # Fallback
            data = np.frombuffer(payload, dtype="<u2")

        if not squeeze:
            data = data.reshape((self._num_sensors, -1))

        return data

    @property
    def squeeze(self):
        return self._squeeze

    @squeeze.setter
    def squeeze(self, squeeze):
        self._squeeze = squeeze


class SocketClient(BaseClient):
    def __init__(self, host, serial_link=False, override_baudrate=None, **kwargs):
        super().__init__(**kwargs)

        if serial_link:
            self._link = links.ExploreSerialLink(host)
        else:
            self._link = links.SocketLink(host)
        self._protocol = None
        self._override_baudrate = override_baudrate

    def _connect(self):
        info = {}

        self._link.connect()

        self._protocol = JsonProtocolBase(self._link)

        msg = self._protocol.get_version()
        if msg:
            startstr = "server version "
            if not msg.startswith(startstr):
                log.warning("server version unknown")
                return info

            server_version_str = msg[len(startstr) :].strip()
            info.update(decode_version_str(server_version_str))

            self._protocol = JsonProtocolStreamingServer(self._link, self.squeeze)
            info["board_sensor_count"] = self._protocol.get_sensor_count()
        else:
            self._protocol = JsonProtocolExplorationServer(self._link, self.squeeze)
            system_info = self._protocol.get_system_info()
            info.update(decode_version_str(system_info["rss_version"]))
            info["sensor"] = system_info["sensor"]
            info["board_sensor_count"] = system_info["sensor_count"]
            info["hw"] = system_info["hw"]

            # Set new baudrate if the link is of ExploreSerialLink type and
            # the baudrate is overriden with a value > 0
            # if the server reports max_baudrate > 0
            baudrate = self._override_baudrate or system_info.get("max_baudrate")
            if isinstance(self._link, links.ExploreSerialLink) and baudrate:
                self._protocol._set_baudrate(baudrate)

        return info

    def _setup_session(self, config):
        cmd = self._protocol.setup_session(config)
        info = self._init_session()

        if "update_rate" in cmd:
            self._link.timeout = 1 / cmd["update_rate"] + self._link.DEFAULT_TIMEOUT
        else:
            self._link.timeout = self._link.DEFAULT_TIMEOUT

        log.debug("setup session")

        return info

    def _start_session(self):
        self._protocol.start_session()

    def _get_next(self):
        return self._protocol.get_next()

    def _stop_session(self):
        self._protocol.stop_session()

    def _disconnect(self):
        self._link.disconnect()

        log.debug("disconnected")

    def _init_session(self, retry=True):
        return self._protocol.init_session()

    @property
    def squeeze(self):
        return self._squeeze

    @squeeze.setter
    def squeeze(self, squeeze):
        if self._protocol:
            self._protocol.squeeze = squeeze
        self._squeeze = squeeze

    @property
    def description(self):
        if isinstance(self._link, links.ExploreSerialLink):
            return f"UART ({self._link._port})"
        else:
            return f"socket ({self._link._host})"


CONFIG_TO_CMD_KEY_MAP = {
    "sensor": "sensors",
    "range_start": "range_start",
    "range_length": "range_length",
    "repetition_mode": "repetition_mode",
    "update_rate": "update_rate",
    "gain": "gain",
    "hw_accelerated_average_samples": "hw_accelerated_average_samples",
    "profile": "profile",
    "bin_count": "bin_count",
    "downsampling_factor": "downsampling_factor",
    "noise_level_normalization": "noise_level_normalization",
    "maximize_signal_attenuation": "maximize_signal_attenuation",
    "running_average_factor": "running_average_factor",
    "sampling_mode": "sampling_mode",
    "sweeps_per_frame": "sweeps_per_frame",
    "sweep_rate": "sweep_rate",
    "tx_disable": "tx_disable",
    "power_save_mode": "power_save_mode",
    "depth_lowpass_cutoff_ratio": "depth_lowpass_cutoff_ratio",
    "asynchronous_measurement": "asynchronous_measurement",
    "mur": "mur",
}

SESSION_HEADER_TO_INFO_KEY_REMAP = {
    "status": None,
    "payload_size": None,
    "start_m": "range_start_m",
    "length_m": "range_length_m",
}

DATA_HEADER_TO_INFO_KEY_REMAP = {}



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

from acconeer.exptool.a111._clients.base import BaseClient, ClientError


class MultiClientWrapper(BaseClient):
    def __init__(self, clients, **kwargs):
        kwargs["squeeze"] = False
        super().__init__(**kwargs)
        self.clients = clients

        for client in clients:
            client.squeeze = False

    def _connect(self):
        for client in self.clients:
            info = client.connect()

        return info

    def _setup_session(self, config):
        expected_sensors = [i + 1 for i in range(len(self.clients))]

        if config.sensor != expected_sensors:
            raise ClientError("Invalid sensor selection for multi client wrapper")

        config.sensor = 1

        for client in self.clients:
            info = client.setup_session(config)

        config.sensor = expected_sensors

        return info

    def _start_session(self):
        for client in self.clients:
            client.start_session()

    def _get_next(self):
        all_info = []
        all_data = []
        for client in self.clients:
            info, data = client.get_next()
            all_info.extend(info)
            all_data.append(data)

        return all_info, np.concatenate(all_data)

    def _stop_session(self):
        for client in self.clients:
            client.stop_session()

    def _disconnect(self):
        for client in self.clients:
            client.disconnect()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

import acconeer.exptool as et
from acconeer.exptool.a111.algo import presence_detection_sparse


def get_sensor_config():
    config = et.a111.SparseServiceConfig()
    config.profile = et.a111.SparseServiceConfig.Profile.PROFILE_2
    config.sampling_mode = et.a111.SparseServiceConfig.SamplingMode.B
    config.range_interval = [0.12, 0.3]
    config.update_rate = 80
    config.sweeps_per_frame = 32
    config.hw_accelerated_average_samples = 60
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):

    detection_threshold = et.configbase.FloatParameter(
        label="Detection threshold",
        default_value=1.4,
        limits=(1.1, 10.0),
        updateable=True,
        order=0,
        help="Level at which to trigger a button press.",
    )

    cool_down_threshold = et.configbase.FloatParameter(
        label="Cool down threshold",
        default_value=1.1,
        limits=(1.0, 5.0),
        updateable=True,
        order=1,
        help="Level at which to be able to trigger again.",
    )

    cool_down_time = et.configbase.FloatParameter(
        label="Cool down time",
        default_value=0,
        limits=(0, 1000),
        updateable=True,
        order=2,
        unit="ms",
        help="Minimal time between triggers.",
    )

    history_length_s = et.configbase.FloatParameter(
        label="Length of display history",
        default_value=10,
        limits=(0, 20),
        updateable=False,
        order=3,
        unit="s",
        help="How long back the display history shows.",
    )

    history_plot_ceiling = et.configbase.FloatParameter(
        label="Ceiling for the trigger value plot",
        default_value=5.0,
        limits=(0, 10.0),
        updateable=True,
        order=1,
        help="Max level for the trigger value plot.",
    )

    def check_sensor_config(self, conf):
        alerts = {
            "processing": [],
            "sensor": [],
        }
        if conf.range_interval[0] < 0.12:
            alerts["sensor"].append(
                et.configbase.Error(
                    "range_interval",
                    "Must be above 0.12. For closer range use Button Press instead.",
                )
            )

        return alerts


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):

        # Config of the presence detector
        presence_config = presence_detection_sparse.ProcessingConfiguration()
        presence_config.detection_threshold = processing_config.detection_threshold
        presence_config.intra_frame_weight = 1.0
        presence_config.intra_frame_time_const = 0.05
        presence_config.output_time_const = 0.02

        self.presence_detector = presence_detection_sparse.Processor(
            sensor_config, presence_config, session_info
        )
        self.processing_config = processing_config
        self.update_rate = sensor_config.update_rate

        history_length = int(processing_config.history_length_s * sensor_config.update_rate)
        self.detection_history = np.zeros(history_length)

        self.cool_trig = True
        self.cool_time = True
        self.cool_counter = 0

        self.update_processing_config(processing_config)

    def update_processing_config(self, processing_config):
        self.cool_down_n_tics = int(processing_config.cool_down_time * self.update_rate / 1000.0)
        self.cool_down_threshold = processing_config.cool_down_threshold
        self.detection_threshold = processing_config.detection_threshold

    def process(self, data, data_info):
        presence_result = self.presence_detector.process(data, data_info)

        score = presence_result["presence_score"]
        button_press = False

        if presence_result["presence_detected"] and self.cool_trig and self.cool_time:
            button_press = True
            self.cool_trig = False
            self.cool_time = False

        if not self.cool_trig and score < self.cool_down_threshold:
            self.cool_trig = True
            self.cool_counter = self.cool_down_n_tics

        if not self.cool_time and self.cool_trig:
            if self.cool_counter > 0:
                self.cool_counter -= 1
            if self.cool_counter == 0:
                self.cool_time = True

        presence_result["detection"] = button_press
        self.detection_history = np.roll(self.detection_history, -1)
        self.detection_history[-1] = button_press
        presence_result["detection_history"] = self.detection_history

        return presence_result



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import configparser
import fnmatch
import subprocess
import sys


def main():
    config = configparser.ConfigParser()
    config.read("setup.cfg")
    section = "check_permissions"

    default_mode = config.get(section, "default_mode", fallback=None)
    overrides = config.get(section, "overrides", fallback="")

    overrides = [s.strip() for s in overrides.split(",")]
    overrides = [s for s in overrides if s]
    overrides = [s.split(":") for s in overrides]

    if not all([len(s) == 2 for s in overrides]):
        raise Exception

    overrides = dict(overrides)

    p = subprocess.run(
        ["git", "ls-files", "--stage"],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        check=True,
    )

    lines = p.stdout.decode().splitlines()
    failed = False
    for line in lines:
        git_mode, _, _, filename = line.split()
        mode = git_mode[-3:]

        try:
            expected_mode = next(m for p, m in overrides.items() if fnmatch.fnmatch(filename, p))
        except StopIteration:
            expected_mode = default_mode

        if not expected_mode:
            continue

        mask = 0o111  # only check execute bit
        if (int(mode, 8) & mask) != (int(expected_mode, 8) & mask):
            print("{}: has mode {}, expected {}".format(filename, mode, expected_mode))
            failed = True

    if failed:
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from ._backend import Backend, ClosedTask
from ._backend_plugin import BackendPlugin
from ._message import (
    BackendPluginStateMessage,
    ConnectionStateMessage,
    GeneralMessage,
    Message,
    PluginStateMessage,
    StatusMessage,
)
from ._model import is_task



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import importlib.resources
from typing import Optional

import qtawesome as qta

from PySide6 import QtCore
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import (
    QButtonGroup,
    QFrame,
    QGraphicsOpacityEffect,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

import pyqtgraph as pg

from acconeer.exptool.app import resources  # type: ignore[attr-defined]
from acconeer.exptool.app.new._enums import PluginFamily
from acconeer.exptool.app.new.app_model import AppModel, PlotPlugin, Plugin, ViewPlugin


class PluginSelectionButton(QPushButton):
    plugin: Plugin

    def __init__(self, plugin: Plugin, parent: QWidget) -> None:
        super().__init__(parent)

        self.plugin = plugin

        self.setText(plugin.title)
        self.setStyleSheet("text-align: left; font-weight: bold;")
        self.setCheckable(True)


class PluginSelectionButtonGroup(QButtonGroup):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setExclusive(True)

    def addButton(self, button: PluginSelectionButton) -> None:
        super().addButton(button)

    def checkedButton(self) -> PluginSelectionButton:
        button = super().checkedButton()
        assert isinstance(button, PluginSelectionButton)
        return button

    def buttons(self) -> list[PluginSelectionButton]:
        buttons = super().buttons()
        assert isinstance(buttons, list)
        assert all(isinstance(e, PluginSelectionButton) for e in buttons)
        return buttons


class PluginSelection(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        app_model.sig_notify.connect(self._on_app_model_update)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(11)

        group_boxes = {}
        for family in PluginFamily:
            group_box = QGroupBox(self)
            group_box.setTitle(family.value)
            group_box.setHidden(True)
            group_box.setLayout(QVBoxLayout(group_box))
            self.layout().addWidget(group_box)
            group_boxes[family] = group_box

        self.button_group = PluginSelectionButtonGroup(self)
        self.button_group.buttonClicked.connect(self._on_load_click)

        for plugin in app_model.plugins:
            group_box = group_boxes[plugin.family]
            group_box.setHidden(False)

            button = PluginSelectionButton(plugin, group_box)
            self.button_group.addButton(button)
            group_box.layout().addWidget(button)

            if plugin.description:
                label = QLabel(group_box)
                label.setText(plugin.description)
                label.setWordWrap(True)
                group_box.layout().addWidget(label)

        self.unload_button = QPushButton("Deselect", self)
        self.unload_button.setStyleSheet("text-align: left;")
        self.unload_button.setFlat(True)
        self.unload_button.clicked.connect(self._on_unload_click)
        self.layout().addWidget(self.unload_button)

    def _on_load_click(self):
        plugin = self.button_group.checkedButton().plugin
        self.app_model.load_plugin(plugin)

    def _on_unload_click(self):
        self.app_model.load_plugin(None)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        plugin: Optional[Plugin] = app_model.plugin

        if plugin is None:
            self.button_group.setExclusive(False)

            for button in self.button_group.buttons():
                button.setChecked(False)

            self.button_group.setExclusive(True)
        else:
            buttons = self.button_group.buttons()
            button = next(b for b in buttons if b.plugin == plugin)
            button.setChecked(True)

        self.unload_button.setEnabled(plugin is not None)

        self.setEnabled(app_model.plugin_state.is_steady)


class PluginPlotArea(QFrame):
    _FPS = 60

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.child_widget: Optional[QWidget] = None
        self.plot_plugin: Optional[PlotPlugin] = None

        self.setObjectName("PluginPlotArea")
        self.setStyleSheet("QFrame#PluginPlotArea {background: #fff; border: 0;}")
        self.setFrameStyle(0)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self.startTimer(int(1000 / self._FPS))

        app_model.sig_load_plugin.connect(self._on_app_model_load_plugin)
        self._on_app_model_load_plugin(app_model.plugin)

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        if self.plot_plugin is None:
            return

        self.plot_plugin.draw()

    def _on_app_model_load_plugin(self, plugin: Optional[Plugin]) -> None:
        if self.plot_plugin is not None:
            # TODO: teardown
            self.plot_plugin = None

        if self.child_widget is not None:
            self.layout().removeWidget(self.child_widget)
            self.child_widget.deleteLater()
            self.child_widget = None

        if plugin is not None:
            self.child_widget = pg.GraphicsLayoutWidget(self)
            self.plot_plugin = plugin.plot_plugin(
                app_model=self.app_model,
                plot_layout=self.child_widget.ci,
            )
        else:
            self.child_widget = PlotPlaceholder(self.app_model, self)

        self.layout().addWidget(self.child_widget)


class PlotPlaceholder(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QHBoxLayout(self))

        self.layout().addStretch(1)

        icon_widget = qta.IconWidget()
        icon_widget.setIconSize(QtCore.QSize(36, 36))
        icon_widget.setIcon(qta.icon("ph.arrow-left-bold", color="#4d5157"))
        self.layout().addWidget(icon_widget)

        label = QLabel("Select a module to begin", self)
        label.setStyleSheet("font-size: 20px;")
        label.setAlignment(QtCore.Qt.AlignCenter)
        self.layout().addWidget(label)

        self.layout().addStretch(1)


class PluginControlArea(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.child_widget: Optional[QWidget] = None
        self.view_plugin: Optional[ViewPlugin] = None

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        app_model.sig_load_plugin.connect(self._on_app_model_load_plugin)
        self._on_app_model_load_plugin(app_model.plugin)

    def _on_app_model_load_plugin(self, plugin: Optional[Plugin]) -> None:
        if self.view_plugin is not None:
            # TODO: teardown
            self.view_plugin = None

        if self.child_widget is not None:
            self.layout().removeWidget(self.child_widget)
            self.child_widget.deleteLater()
            self.child_widget = None

        if plugin is not None:
            self.child_widget = QWidget(self)
            self.view_plugin = plugin.view_plugin(
                app_model=self.app_model,
                view_widget=self.child_widget,
            )
        else:
            self.child_widget = ControlPlaceholder(self.app_model, self)

        self.layout().addWidget(self.child_widget)


class ControlPlaceholder(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QGridLayout(self))

        with importlib.resources.path(resources, "icon-black.svg") as path:
            icon = QSvgWidget(str(path), self)

        icon.setMaximumSize(60, 60)
        icon.renderer().setAspectRatioMode(QtCore.Qt.KeepAspectRatio)
        effect = QGraphicsOpacityEffect(icon)
        effect.setOpacity(0.1)
        icon.setGraphicsEffect(effect)

        self.layout().addWidget(icon)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.distance import (
    Detector,
    DetectorConfig,
    DetectorResult,
    ThresholdMethod,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()
    detector_config = DetectorConfig(
        start_m=0.0,
        end_m=2.0,
        max_profile=a121.Profile.PROFILE_3,
        max_step_length=12,
        threshold_method=ThresholdMethod.RECORDED,
    )
    detector = Detector(client=client, sensor_id=1, detector_config=detector_config)

    detector.calibrate_close_range()

    detector.record_threshold()

    detector.start()

    pg_updater = PGUpdater(num_curves=len(detector.processor_specs))
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        detector_result = detector.get_next()
        try:
            pg_process.put_data(detector_result)
        except et.PGProccessDiedException:
            break

    detector.stop()

    print("Disconnecting...")
    client.disconnect()


class PGUpdater:
    def __init__(self, num_curves):
        self.num_curves = num_curves
        self.distance_history = [np.NaN] * 100

    def setup(self, win):
        self.sweep_plot = win.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.addItem(pg.PlotDataItem())

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.sweep_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        pen = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.threshold_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        self.dist_history_plot = win.addPlot(row=1, col=0)
        self.dist_history_plot.setMenuEnabled(False)
        self.dist_history_plot.showGrid(x=True, y=True)
        self.dist_history_plot.addLegend()
        self.dist_history_plot.setLabel("left", "Estimated_distance")
        self.dist_history_plot.addItem(pg.PlotDataItem())

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.dist_history_curve = self.dist_history_plot.plot(**feat_kw)

    def update(self, result: DetectorResult):
        self.distance_history.pop(0)
        self.distance_history.append(result.distances[0])

        for idx, processor_result in enumerate(result.processor_results):
            threshold = processor_result.extra_result.used_threshold
            valid_threshold_idx = np.where(~np.isnan(threshold))[0]
            threshold = threshold[valid_threshold_idx]
            self.sweep_curves[idx].setData(
                processor_result.extra_result.distances_m, processor_result.extra_result.abs_sweep
            )
            self.threshold_curves[idx].setData(
                processor_result.extra_result.distances_m[valid_threshold_idx], threshold
            )
        self.dist_history_curve.setData(self.distance_history)


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import logging
import pickle
from pathlib import Path
from typing import Any, Callable, Optional

import attrs
import numpy as np
import qtawesome as qta

from PySide6.QtWidgets import QPushButton, QVBoxLayout, QWidget

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._plugins import (
    DetectorBackendPluginBase,
    DetectorPlotPluginBase,
    DetectorViewPluginBase,
)
from acconeer.exptool.app.new import (
    BUTTON_ICON_COLOR,
    AppModel,
    ConnectionState,
    GeneralMessage,
    HandledException,
    Message,
    Plugin,
    PluginFamily,
    PluginGeneration,
    PluginState,
    PluginStateMessage,
    get_temp_h5_path,
    is_task,
)
from acconeer.exptool.app.new.ui.plugin import (
    AttrsConfigEditor,
    GridGroupBox,
    PidgetFactoryMapping,
    pidgets,
    utils,
)

from ._detector import Detector, DetectorConfig, DetectorResult, _load_algo_data


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: DetectorConfig = attrs.field(factory=DetectorConfig)
    replaying: bool = attrs.field(default=False)


@attrs.frozen(kw_only=True)
class Save:
    config: DetectorConfig = attrs.field()


class BackendPlugin(DetectorBackendPluginBase[SharedState]):
    def __init__(self, callback: Callable[[Message], None], key: str) -> None:
        super().__init__(callback=callback, key=key)

        self._started: bool = False
        self._live_client: Optional[a121.Client] = None
        self._replaying_client: Optional[a121._ReplayingClient] = None
        self._recorder: Optional[a121.H5Recorder] = None
        self._opened_record: Optional[a121.H5Record] = None
        self._detector_instance: Optional[Detector] = None

        self.restore_defaults()

    @is_task
    def deserialize(self, *, data: bytes) -> None:
        try:
            obj = pickle.loads(data)
        except Exception:
            log.warning("Could not load pickled - pickle.loads() failed")
            return

        if not isinstance(obj, Save):
            log.warning("Could not load pickled - not the correct type")
            return

        if not isinstance(obj.config, DetectorConfig):
            log.warning("Could not load pickled - not the correct type")
            return

        self.shared_state.config = obj.config
        self.broadcast(sync=True)

    def _serialize(self) -> bytes:
        obj = Save(config=self.shared_state.config)
        return pickle.dumps(obj, protocol=4)

    def broadcast(self, sync: bool = False) -> None:
        super().broadcast()

        if sync:
            self.callback(GeneralMessage(name="sync", recipient="view_plugin"))

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast(sync=True)

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast(sync=True)

    @property
    def _client(self) -> Optional[a121.Client]:
        if self._replaying_client is not None:
            return self._replaying_client

        return self._live_client

    def idle(self) -> bool:
        if self._started:
            self._get_next()
            return True
        else:
            return False

    def attach_client(self, *, client: Any) -> None:
        self._live_client = client

    def detach_client(self) -> None:
        self._live_client = None

    @is_task
    def update_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    def teardown(self) -> None:
        self.callback(
            GeneralMessage(
                name="serialized",
                kwargs={
                    "generation": PluginGeneration.A121,
                    "key": self.key,
                    "data": self._serialize(),
                },
            )
        )
        self.detach_client()

    @is_task
    def load_from_file(self, *, path: Path) -> None:
        try:
            self._load_from_file_setup(path=path)
        except Exception as exc:
            self._opened_record = None
            self._replaying_client = None
            self.shared_state.replaying = False

            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            raise HandledException("Could not load from file") from exc

        self.start_session(with_recorder=False)

        self.shared_state.replaying = True

        self.send_status_message(f"<b>Replaying from {path.name}</b>")
        self.broadcast(sync=True)

    def _load_from_file_setup(self, *, path: Path) -> None:
        r = a121.open_record(path)
        assert isinstance(r, a121.H5Record)
        self._opened_record = r
        self._replaying_client = a121._ReplayingClient(self._opened_record)

        algo_group = self._opened_record.get_algo_group(self.key)
        _, config = _load_algo_data(algo_group)
        self.shared_state.config = config

    @is_task
    def start_session(self, *, with_recorder: bool = True) -> None:
        if self._started:
            raise RuntimeError

        if self._client is None:
            raise RuntimeError

        if not self._client.connected:
            raise RuntimeError

        self._detector_instance = Detector(
            client=self._client,
            sensor_id=self.shared_state.sensor_id,
            detector_config=self.shared_state.config,
        )

        if with_recorder:
            self._recorder = a121.H5Recorder(get_temp_h5_path())
        else:
            self._recorder = None

        try:
            self._detector_instance.start(self._recorder)
        except Exception as exc:
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            raise HandledException("Could not start") from exc

        self._started = True

        self.broadcast()

        self.callback(
            GeneralMessage(
                name="setup",
                kwargs=dict(
                    detector_config=self.shared_state.config,
                    sensor_config=Detector._get_sensor_config(self.shared_state.config),
                ),
                recipient="plot_plugin",
            )
        )
        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

    @is_task
    def stop_session(self) -> None:
        if not self._started:
            raise RuntimeError

        if self._detector_instance is None:
            raise RuntimeError

        try:
            self._detector_instance.stop()
        except Exception as exc:
            raise HandledException("Failure when stopping session") from exc
        finally:
            if self._recorder is not None:
                assert self._recorder.path is not None
                path = Path(self._recorder.path)
                self.callback(GeneralMessage(name="saveable_file", data=path))
                self._recorder = None

            if self.shared_state.replaying:
                assert self._opened_record is not None
                self._opened_record.close()

                self._opened_record = None
                self._replaying_client = None

                self.shared_state.replaying = False

            self._started = False
            self.broadcast()
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            self.callback(GeneralMessage(name="result_tick_time", data=None))

    def _get_next(self) -> None:
        if not self._started:
            raise RuntimeError

        if self._detector_instance is None:
            raise RuntimeError

        try:
            result = self._detector_instance.get_next()
        except a121._StopReplay:
            self.stop_session()
            return
        except Exception as exc:
            try:
                self.stop_session()
            except Exception:
                pass

            raise HandledException("Failed to get_next") from exc

        self.callback(
            GeneralMessage(name="result_tick_time", data=result.service_result.tick_time)
        )

        self.callback(GeneralMessage(name="plot", data=result, recipient="plot_plugin"))


class PlotPlugin(DetectorPlotPluginBase):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)

    def setup_from_message(self, message: GeneralMessage) -> None:
        assert message.kwargs is not None
        self.setup(**message.kwargs)

    def update_from_message(self, message: GeneralMessage) -> None:
        assert isinstance(message.data, DetectorResult)
        self.update(message.data)

    def setup(self, detector_config: DetectorConfig, sensor_config: a121.SensorConfig) -> None:
        self.detector_config = detector_config
        self.distances = np.linspace(
            detector_config.start_m, detector_config.end_m, sensor_config.num_points
        )

        self.history_length_s = 10

        max_num_of_sectors = max(6, self.distances.size // 3)
        self.sector_size = max(1, -(-self.distances.size // max_num_of_sectors))
        self.num_sectors = -(-self.distances.size // self.sector_size)
        self.sector_offset = (self.num_sectors * self.sector_size - self.distances.size) // 2

        win = self.plot_layout

        self.limit_lines = []
        dashed_pen = pg.mkPen("k", width=2.5)  # , style=QtCore.Qt.DashLine

        # Amplitude plot

        self.ampl_plot = win.addPlot(
            row=0,
            col=0,
            title="Amplitude, sweeps (orange), mean sweep (blue)",
        )

        self.ampl_plot.setMenuEnabled(False)
        self.ampl_plot.setMouseEnabled(x=False, y=False)
        self.ampl_plot.hideButtons()
        self.ampl_plot.showGrid(x=True, y=True)
        self.ampl_plot.setLabel("bottom", "Distance (m)")
        self.ampl_plot.setLabel("left", "Amplitude")

        self.ampl_plot.setYRange(0, 2**16)

        self.frame_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(1),
        )
        self.mean_sweep_scatter = pg.ScatterPlotItem(
            size=10,
            brush=et.utils.pg_brush_cycler(0),
        )

        self.ampl_plot.addItem(self.frame_scatter)
        self.ampl_plot.addItem(self.mean_sweep_scatter)
        self.frame_smooth_limits = et.utils.SmoothLimits(self.detector_config.frame_rate)

        # Noise estimation plot

        self.noise_plot = win.addPlot(
            row=1,
            col=0,
            title="Noise",
        )
        self.noise_plot.setMenuEnabled(False)
        self.noise_plot.setMouseEnabled(x=False, y=False)
        self.noise_plot.hideButtons()
        self.noise_plot.showGrid(x=True, y=True)
        self.noise_plot.setLabel("bottom", "Distance (m)")
        self.noise_plot.setLabel("left", "Amplitude")
        self.noise_plot.setVisible(False)
        self.noise_curve = self.noise_plot.plot(pen=et.utils.pg_pen_cycler())
        self.noise_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)

        # Depthwise presence plot

        self.move_plot = win.addPlot(
            row=2,
            col=0,
            title="Depthwise presence",
        )
        self.move_plot.setMenuEnabled(False)
        self.move_plot.setMouseEnabled(x=False, y=False)
        self.move_plot.hideButtons()
        self.move_plot.showGrid(x=True, y=True)
        self.move_plot.setLabel("bottom", "Distance (m)")
        self.move_plot.setLabel("left", "Norm. ampl.")
        zero_curve = self.move_plot.plot(self.distances, np.zeros_like(self.distances))
        self.inter_curve = self.move_plot.plot()
        self.total_curve = self.move_plot.plot()
        self.move_smooth_max = et.utils.SmoothMax(
            self.detector_config.frame_rate,
            tau_decay=1.0,
            tau_grow=0.25,
        )

        self.move_depth_line = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5))
        self.move_depth_line.hide()
        self.move_plot.addItem(self.move_depth_line)
        limit_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.move_plot.addItem(limit_line)
        self.limit_lines.append(limit_line)

        fbi = pg.FillBetweenItem(
            zero_curve,
            self.inter_curve,
            brush=et.utils.pg_brush_cycler(0),
        )
        self.move_plot.addItem(fbi)

        fbi = pg.FillBetweenItem(
            self.inter_curve,
            self.total_curve,
            brush=et.utils.pg_brush_cycler(1),
        )
        self.move_plot.addItem(fbi)

        # Presence history plot

        self.move_hist_plot = pg.PlotItem(title="Presence history")
        self.move_hist_plot.setMenuEnabled(False)
        self.move_hist_plot.setMouseEnabled(x=False, y=False)
        self.move_hist_plot.hideButtons()
        self.move_hist_plot.showGrid(x=True, y=True)
        self.move_hist_plot.setLabel("bottom", "Time (s)")
        self.move_hist_plot.setLabel("left", "Score")
        self.move_hist_plot.setXRange(-self.history_length_s, 0)
        self.history_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)
        self.move_hist_plot.setYRange(0, 10)

        self.move_hist_curve = self.move_hist_plot.plot(pen=et.utils.pg_pen_cycler())
        limit_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.move_hist_plot.addItem(limit_line)
        self.limit_lines.append(limit_line)

        self.present_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )
        not_present_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("No presence detected")
        )
        self.present_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.not_present_text_item = pg.TextItem(
            html=not_present_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.move_hist_plot.addItem(self.present_text_item)
        self.move_hist_plot.addItem(self.not_present_text_item)
        self.present_text_item.hide()
        self.not_present_text_item.hide()

        for line in self.limit_lines:
            line.setPos(self.detector_config.detection_threshold)

        # Sector plot

        self.sector_plot = pg.PlotItem()
        self.sector_plot.setAspectLocked()
        self.sector_plot.hideAxis("left")
        self.sector_plot.hideAxis("bottom")
        self.sectors = []

        pen = pg.mkPen("k", width=1)
        span_deg = 25
        for r in np.flip(np.arange(self.num_sectors) + 1):
            sector = pg.QtGui.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
            sector.setStartAngle(-16 * span_deg)
            sector.setSpanAngle(16 * span_deg * 2)
            sector.setPen(pen)
            self.sector_plot.addItem(sector)
            self.sectors.append(sector)

        self.sectors.reverse()

        sublayout = win.addLayout(row=3, col=0)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.move_hist_plot, col=0)
        sublayout.addItem(self.sector_plot, col=1)

    def update(self, data: DetectorResult) -> None:
        amplitudes = np.abs(data.processor_extra_result.frame)
        self.frame_scatter.setData(
            np.tile(self.distances, self.detector_config.sweeps_per_frame),
            amplitudes.flatten(),
        )

        self.mean_sweep_scatter.setData(self.distances, data.processor_extra_result.mean_sweep)
        self.ampl_plot.setYRange(*self.frame_smooth_limits.update(amplitudes))

        noise = data.processor_extra_result.lp_noise
        self.noise_curve.setData(self.distances, noise)
        self.noise_plot.setYRange(0, self.noise_smooth_max.update(noise))

        movement_x = data.presence_distance

        move_ys = data.processor_extra_result.depthwise_presence
        self.inter_curve.setData(self.distances, data.processor_extra_result.inter)
        self.total_curve.setData(self.distances, move_ys)
        m = self.move_smooth_max.update(np.max(move_ys))
        m = max(m, 2 * self.detector_config.detection_threshold)
        self.move_plot.setYRange(0, m)
        self.move_depth_line.setPos(movement_x)
        self.move_depth_line.setVisible(bool(data.presence_detected))

        move_hist_ys = data.processor_extra_result.presence_history
        move_hist_xs = np.linspace(-self.history_length_s, 0, len(move_hist_ys))

        m_hist = max(float(np.max(move_hist_ys)), self.detector_config.detection_threshold * 1.05)
        m_hist = self.history_smooth_max.update(m_hist)

        self.move_hist_plot.setYRange(0, m_hist)
        self.move_hist_curve.setData(move_hist_xs, move_hist_ys)
        self.set_present_text_y_pos(m_hist)

        if data.presence_detected:
            present_text = "Presence detected at {:.0f} cm".format(movement_x * 100)
            present_html = self.present_html_format.format(present_text)
            self.present_text_item.setHtml(present_html)

            self.present_text_item.show()
            self.not_present_text_item.hide()
        else:
            self.present_text_item.hide()
            self.not_present_text_item.show()

        brush = et.utils.pg_brush_cycler(0)
        for sector in self.sectors:
            sector.setBrush(brush)

        if data.presence_detected:
            index = (
                data.processor_extra_result.presence_distance_index + self.sector_offset
            ) // self.sector_size
            self.sectors[index].setBrush(et.utils.pg_brush_cycler(1))

    def set_present_text_y_pos(self, y):
        self.present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)
        self.not_present_text_item.setPos(-self.history_length_s / 2, 0.95 * y)


class ViewPlugin(DetectorViewPluginBase):
    def __init__(self, app_model: AppModel, view_widget: QWidget) -> None:
        super().__init__(app_model=app_model, view_widget=view_widget)

        self.view_layout = QVBoxLayout(self.view_widget)
        self.view_layout.setContentsMargins(0, 0, 0, 0)
        self.view_widget.setLayout(self.view_layout)

        # TODO: Fix parents

        self.start_button = QPushButton(
            qta.icon("fa5s.play-circle", color=BUTTON_ICON_COLOR),
            "Start measurement",
            self.view_widget,
        )
        self.start_button.setShortcut("space")
        self.start_button.clicked.connect(self._send_start_request)
        self.stop_button = QPushButton(
            qta.icon("fa5s.stop-circle", color=BUTTON_ICON_COLOR),
            "Stop",
            self.view_widget,
        )
        self.stop_button.setShortcut("space")
        self.stop_button.clicked.connect(self._send_stop_request)

        self.defaults_button = QPushButton(
            qta.icon("mdi6.restore", color=BUTTON_ICON_COLOR),
            "Reset settings and calibrations",
            self.view_widget,
        )
        self.defaults_button.clicked.connect(self._send_defaults_request)

        button_group = GridGroupBox("Controls", parent=self.view_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.defaults_button, 1, 0, 1, -1)
        self.view_layout.addWidget(button_group)

        sensor_selection_group = utils.VerticalGroupBox(
            "Sensor selection", parent=self.view_widget
        )
        self.sensor_id_pidget = pidgets.SensorIdParameterWidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_parameter_changed.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        self.view_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor[DetectorConfig](
            title="Detector parameters",
            factory_mapping=self._get_pidget_mapping(),
            parent=self.view_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        self.view_layout.addWidget(self.config_editor)

        self.view_layout.addStretch(1)

    @classmethod
    def _get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "start_m": pidgets.FloatParameterWidgetFactory(
                name_label_text="Range start",
                suffix=" m",
                decimals=3,
            ),
            "end_m": pidgets.FloatParameterWidgetFactory(
                name_label_text="Range end",
                suffix=" m",
                decimals=3,
            ),
            "step_length": pidgets.OptionalIntParameterWidgetFactory(
                name_label_text="Step length",
                checkbox_label_text="Override",
                limits=(1, None),
                init_set_value=24,
            ),
            "detection_threshold": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Detection threshold",
                decimals=2,
                limits=(0, 5),
            ),
            "frame_rate": pidgets.FloatParameterWidgetFactory(
                name_label_text="Frame rate",
                suffix=" Hz",
                decimals=1,
                limits=(1, 100),
            ),
            "sweeps_per_frame": pidgets.IntParameterWidgetFactory(
                name_label_text="Sweeps per frame",
                limits=(1, 4095),
            ),
            "hwaas": pidgets.IntParameterWidgetFactory(
                name_label_text="HWAAS",
                limits=(1, 511),
            ),
        }

    def on_app_model_update(self, app_model: AppModel) -> None:
        state = app_model.backend_plugin_state
        self.sensor_id_pidget.update_available_sensor_list(app_model._a121_server_info)

        if state is None:
            self.start_button.setEnabled(False)
            self.stop_button.setEnabled(False)
            self.defaults_button.setEnabled(False)

            self.config_editor.set_data(None)
            self.config_editor.setEnabled(False)
            self.sensor_id_pidget.set_parameter(None)

            return

        assert isinstance(state, SharedState)

        self.defaults_button.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)

        self.config_editor.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)
        self.config_editor.set_data(state.config)
        self.sensor_id_pidget.set_parameter(state.sensor_id)
        self.sensor_id_pidget.setEnabled(app_model.plugin_state.is_steady)

        ready_for_session = (
            app_model.plugin_state == PluginState.LOADED_IDLE
            and app_model.connection_state == ConnectionState.CONNECTED
        )

        self.start_button.setEnabled(ready_for_session)
        self.stop_button.setEnabled(app_model.plugin_state == PluginState.LOADED_BUSY)

    # TODO: move to detector base (?)
    def _on_config_update(self, config: DetectorConfig) -> None:
        self.app_model.put_backend_plugin_task("update_config", {"config": config})

    # TODO: move to detector base (?)
    def handle_message(self, message: GeneralMessage) -> None:
        if message.name == "sync":
            log.debug(f"{type(self).__name__} syncing")

            self.config_editor.sync()
        else:
            raise RuntimeError("Unknown message")

    # TODO: move to detector base (?)
    def _send_start_request(self) -> None:
        self.app_model.put_backend_plugin_task("start_session", on_error=self.app_model.emit_error)
        self.app_model.set_plugin_state(PluginState.LOADED_STARTING)

    # TODO: move to detector base (?)
    def _send_stop_request(self) -> None:
        self.app_model.put_backend_plugin_task("stop_session", on_error=self.app_model.emit_error)
        self.app_model.set_plugin_state(PluginState.LOADED_STOPPING)

    def _send_defaults_request(self) -> None:
        self.app_model.put_backend_plugin_task("restore_defaults")

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        self.app_model.put_backend_plugin_task("update_sensor_id", {"sensor_id": sensor_id})

    # TODO: move to detector base (?)
    def teardown(self) -> None:
        self.view_layout.deleteLater()


PRESENCE_DETECTOR_PLUGIN = Plugin(
    generation=PluginGeneration.A121,
    key="presence_detector",
    title="Presence detector",
    # description="",
    family=PluginFamily.DETECTOR,
    backend_plugin=BackendPlugin,
    plot_plugin=PlotPlugin,
    view_plugin=ViewPlugin,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

# THIS FILE IS AUTOMATICALLY GENERATED - DO NOT EDIT

import logging
import queue

import serial

from acconeer.exptool.flash._xc120._uart_protocol import Packet, UartReader


_LOG = logging.getLogger(__name__)


class CommandFailed(Exception):
    pass


class XcCommandResponsePacket(Packet):
    packet_type = 0xF2

    def __init__(self, payload):
        self.command_id = int.from_bytes(payload[0:2], byteorder="little")
        self.command_payload = payload[2:]
        super().__init__(payload)

    def get_command_packet(self):
        return _command_packets.get(self.command_id)(self.command_payload)


class XcCommandRequestPacket(Packet):
    packet_type = 0xF2

    def __init__(self, command_payload):
        command_payload[0:0] = self.command_id.to_bytes(2, byteorder="little")
        super().__init__(command_payload)


class GetLastErrorRequestPacket(XcCommandRequestPacket):
    command_id = 0x0002

    def __init__(
        self,
    ):
        payload = bytearray()
        super().__init__(payload)


class GetLastErrorResponsePacket:
    command_id = 0x0002

    def __init__(self, payload):
        self.payload = payload

    def get_status(self):
        return self.payload[0]

    def get_response_data(self):
        return self.payload[1:].decode("ascii")


class GetAppSwVersionRequestPacket(XcCommandRequestPacket):
    command_id = 0x010A

    def __init__(
        self,
    ):
        payload = bytearray()
        super().__init__(payload)


class GetAppSwVersionResponsePacket:
    command_id = 0x010A

    def __init__(self, payload):
        self.payload = payload

    def get_status(self):
        return self.payload[0]

    def get_version(self):
        return self.payload[1:].decode("ascii")


class GetAppSwNameRequestPacket(XcCommandRequestPacket):
    command_id = 0x010B

    def __init__(
        self,
    ):
        payload = bytearray()
        super().__init__(payload)


class GetAppSwNameResponsePacket:
    command_id = 0x010B

    def __init__(self, payload):
        self.payload = payload

    def get_status(self):
        return self.payload[0]

    def get_name(self):
        return self.payload[1:].decode("ascii")


class DfuRebootRequestPacket(XcCommandRequestPacket):
    command_id = 0xFFFF

    def __init__(
        self,
    ):
        payload = bytearray()
        super().__init__(payload)


_command_packets = {
    0x0002: GetLastErrorResponsePacket,
    0x010A: GetAppSwVersionResponsePacket,
    0x010B: GetAppSwNameResponsePacket,
}


class XCCommunication:
    def __init__(self, port):
        self._ser = serial.Serial(port, exclusive=True)
        self._reader = UartReader(self._ser, [XcCommandResponsePacket])
        self._reader.start()
        self._timeout = UartReader.DEFAULT_READ_TIMEOUT
        self._reader_started = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def stop(self):
        if self._reader_started:
            self._reader.stop()
            self._reader_started = False

    def close(self):
        self.stop()
        self._ser.close()
        self._ser = None

    def _send_packet(self, packet):
        data = packet.get_byte_array()
        _LOG.debug("Sending packet %s", data)
        self._ser.write(data)

    def _execute_request(self, packet):
        try:
            self._send_packet(packet)
            response = self._reader.wait_packet(XcCommandResponsePacket.packet_type, self._timeout)
            return response.get_command_packet()
        except queue.Empty:
            _LOG.error("read timeout")
            raise TimeoutError("_execute_request timeout")

    def get_last_error(self):
        response = self._execute_request(GetLastErrorRequestPacket())
        status = response.get_status()
        response_data = response.get_response_data()
        if status != 0:
            error_msg = self.get_last_error()
            raise CommandFailed(f"Command failed with code {status} [{error_msg}]")
        return response_data

    def get_app_sw_version(self):
        response = self._execute_request(GetAppSwVersionRequestPacket())
        status = response.get_status()
        version = response.get_version()
        if status != 0:
            error_msg = self.get_last_error()
            raise CommandFailed(f"Command failed with code {status} [{error_msg}]")
        return version

    def get_app_sw_name(self):
        response = self._execute_request(GetAppSwNameRequestPacket())
        status = response.get_status()
        name = response.get_name()
        if status != 0:
            error_msg = self.get_last_error()
            raise CommandFailed(f"Command failed with code {status} [{error_msg}]")
        return name

    def dfu_reboot(self):
        self._send_packet(DfuRebootRequestPacket())



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import numpy as np

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.distance import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorMode,
    ThresholdMethod,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()

    sensor_config = a121.SensorConfig(
        subsweeps=[
            a121.SubsweepConfig(
                start_point=50,
                step_length=4,
                num_points=25,
                profile=a121.Profile.PROFILE_1,
                hwaas=10,
                phase_enhancement=True,
            ),
            a121.SubsweepConfig(
                start_point=150,
                step_length=4,
                num_points=25,
                profile=a121.Profile.PROFILE_1,
                hwaas=20,
                phase_enhancement=True,
            ),
        ],
        sweeps_per_frame=5,
    )

    metadata = client.setup_session(sensor_config)

    threshold_config = ProcessorConfig(
        processor_mode=ProcessorMode.RECORDED_THRESHOLD_CALIBRATION,
        threshold_method=ThresholdMethod.RECORDED,
    )
    threshold_processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=threshold_config,
    )

    pg_updater = PGUpdater(None)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    sc_bg_num_sweeps = 20

    for _ in range(sc_bg_num_sweeps):
        result = client.get_next()
        processed_data = threshold_processor.process(result)

    distance_context = ProcessorContext(recorded_threshold=processed_data.recorded_threshold)
    distance_config = ProcessorConfig(
        processor_mode=ProcessorMode.DISTANCE_ESTIMATION, threshold_method=ThresholdMethod.CFAR
    )
    distance_processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=distance_config,
        context=distance_context,
    )

    while not interrupt_handler.got_signal:
        extended_result = client.get_next()
        processed_data = distance_processor.process(extended_result)
        try:
            pg_process.put_data(processed_data)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    pg_process.close()
    client.disconnect()


class PGUpdater:
    def __init__(self, processing_config):
        self.processing_config = processing_config

    def setup(self, win):
        self.sweep_plot = win.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.addItem(pg.PlotDataItem())

        legends = ["Sweep", "Threshold"]
        self.curves = {}
        for i, legend in enumerate(legends):
            pen = et.utils.pg_pen_cycler(i)
            brush = et.utils.pg_brush_cycler(i)
            symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
            feat_kw = dict(pen=pen, **symbol_kw)
            self.curves[legend] = self.sweep_plot.plot(**feat_kw, name=legends[i])

        self.smooth_max = et.utils.SmoothMax()

    def update(self, d):
        threshold = d.extra_result.used_threshold
        valid_threshold_idx = np.where(~np.isnan(threshold))[0]
        threshold = threshold[valid_threshold_idx]

        self.curves["Sweep"].setData(d.extra_result.abs_sweep)
        self.curves["Threshold"].setData(valid_threshold_idx, threshold)
        self.sweep_plot.setYRange(
            0,
            self.smooth_max.update(np.amax(np.concatenate((d.extra_result.abs_sweep, threshold)))),
        )


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from . import utils
from .platform_install import PlatformInstall
from .setup_group import SetupGroup
from .setup_step import RequireFileContentStep, SetupStep, ShellCommandStep



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .agnostic_client import AgnosticClient, ClientError
from .communication_protocol import CommunicationProtocol
from .link import BufferedLink, Link
from .recorder import Recorder



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt
from numpy import cos, pi, sqrt, square
from scipy.special import binom

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import AlgoConfigBase, ProcessorBase
from acconeer.exptool.a121.algo._utils import get_distances_m


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoConfigBase):
    detection_threshold: float = attrs.field(default=1.5)
    intra_frame_weight: float = attrs.field(default=0.6)
    inter_frame_fast_cutoff: float = attrs.field(default=20.0)
    inter_frame_slow_cutoff: float = attrs.field(default=0.2)
    inter_frame_deviation_time_const: float = attrs.field(default=0.5)
    intra_frame_time_const: float = attrs.field(default=0.15)
    output_time_const: float = attrs.field(default=0.5)
    history_length_s: int = attrs.field(default=5)


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    frame: npt.NDArray[np.complex_] = attrs.field()
    mean_sweep: npt.NDArray[np.float_] = attrs.field()
    fast_lp_mean_sweep: npt.NDArray[np.float_] = attrs.field()
    slow_lp_mean_sweep: npt.NDArray[np.float_] = attrs.field()
    lp_noise: npt.NDArray[np.float_] = attrs.field()
    inter: npt.NDArray[np.float_] = attrs.field()
    intra: npt.NDArray[np.float_] = attrs.field()
    depthwise_presence: npt.NDArray[np.float_] = attrs.field()
    presence_distance_index: int = attrs.field()
    presence_history: npt.NDArray[np.float_] = attrs.field()


@attrs.frozen(kw_only=True)
class ProcessorResult:
    presence_score: float = attrs.field()
    presence_distance: float = attrs.field()
    presence_detected: bool = attrs.field()
    extra_result: ProcessorExtraResult = attrs.field()


class Processor(ProcessorBase[ProcessorConfig, ProcessorResult]):
    # lp(f): low pass (filtered)
    # cut: cutoff frequency [Hz]
    # tc: time constant [s]
    # sf: smoothing factor [dimensionless]

    ENVELOPE_FWHM_M = {
        a121.Profile.PROFILE_1: 0.04,
        a121.Profile.PROFILE_2: 0.07,
        a121.Profile.PROFILE_3: 0.14,
        a121.Profile.PROFILE_4: 0.19,
        a121.Profile.PROFILE_5: 0.32,
    }

    APPROX_BASE_STEP_LENGTH_M = 2.5e-3

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        subsweep_index: Optional[int] = None,
    ) -> None:
        if subsweep_index is None:
            if len(sensor_config.subsweeps) > 1:
                raise ValueError("Multiple subsweeps are not supported")

            subsweep_index = 0

        self.sensor_config = sensor_config
        self.metadata = metadata
        self.processor_config = processor_config
        self.subsweep_index = subsweep_index

        if self.sensor_config.frame_rate is None:
            raise ValueError("Frame rate must be set")

        self.sweeps_per_frame = self.sensor_config.sweeps_per_frame
        self.distances, _ = get_distances_m(self.sensor_config, metadata)
        self.num_distances = self.distances.size
        self.f = self.sensor_config.frame_rate

        points_per_pulse = self.ENVELOPE_FWHM_M[sensor_config.profile] / (
            self.APPROX_BASE_STEP_LENGTH_M * sensor_config.step_length
        )
        self.depth_filter_length = max(int(round(points_per_pulse)), 1)

        # Fixed parameters
        self.noise_est_diff_order = 3
        noise_tc = 10.0

        if self.sweeps_per_frame <= self.noise_est_diff_order:
            raise ValueError(
                f"Number of sweeps per frame must be greater than {self.noise_est_diff_order}"
            )

        self.noise_sf = self._tc_to_sf(noise_tc, self.f)

        nd = self.noise_est_diff_order
        self.noise_norm_factor = np.sqrt(np.sum(np.square(binom(nd, np.arange(nd + 1)))))

        self.fast_lp_mean_sweep = np.zeros(self.num_distances)
        self.slow_lp_mean_sweep = np.zeros(self.num_distances)
        self.lp_inter_dev = np.zeros(self.num_distances)
        self.lp_intra_dev = np.zeros(self.num_distances)
        self.lp_noise = np.zeros(self.num_distances)

        self.presence_score = 0
        self.presence_distance_index = 0
        self.presence_distance = 0

        self.presence_history = np.zeros(int(round(self.f * processor_config.history_length_s)))
        self.update_index = 0

        self.threshold = processor_config.detection_threshold
        self.intra_weight = processor_config.intra_frame_weight
        self.inter_weight = 1.0 - self.intra_weight

        self.update_config(processor_config)

    def update_config(self, processor_config: ProcessorConfig) -> None:
        self.threshold = processor_config.detection_threshold
        self.intra_weight = processor_config.intra_frame_weight
        self.inter_weight = 1.0 - self.intra_weight

        self.fast_sf = self._cutoff_to_sf(processor_config.inter_frame_fast_cutoff, self.f)
        self.slow_sf = self._cutoff_to_sf(processor_config.inter_frame_slow_cutoff, self.f)
        self.inter_dev_sf = self._tc_to_sf(
            processor_config.inter_frame_deviation_time_const, self.f
        )
        self.intra_sf = self._tc_to_sf(processor_config.intra_frame_time_const, self.f)
        self.output_sf = self._tc_to_sf(processor_config.output_time_const, self.f)

    @staticmethod
    def _cutoff_to_sf(fc: float, fs: float) -> float:
        """Cutoff frequency to smoothing factor conversion"""

        if fc > 0.5 * fs:
            return 0.0

        cos_w = cos(2.0 * pi * (fc / fs))
        return 2.0 - cos_w - sqrt(square(cos_w) - 4.0 * cos_w + 3.0)  # type: ignore[no-any-return]

    @staticmethod
    def _tc_to_sf(tc: float, fs: float) -> float:
        """Time constant to smoothing factor conversion"""

        if tc <= 0.0:
            return 0.0

        return np.exp(-1.0 / (tc * fs))  # type: ignore[no-any-return]

    @staticmethod
    def _dynamic_sf(static_sf: float, update_index: int) -> float:
        return min(static_sf, 1.0 - 1.0 / (1.0 + update_index))

    @staticmethod
    def _abs_dev(
        a: npt.NDArray[np.complex_],
        axis: Optional[int] = None,
        ddof: int = 0,
        subtract_mean: bool = True,
    ) -> npt.NDArray[np.float_]:
        if subtract_mean:
            a = a - a.mean(axis=axis, keepdims=True)

        if axis is None:
            n = a.size
        else:
            n = a.shape[axis]

        if ddof < 0:
            raise ValueError("ddof must be greater than or equal to 0")

        if n <= ddof:
            raise ValueError("n must be greater than ddof")

        return np.mean(np.abs(a), axis=axis) * sqrt(n / (n - ddof))  # type: ignore[no-any-return]

    @staticmethod
    def _depth_filter(a: npt.NDArray, depth_filter_length: int) -> npt.NDArray[np.float_]:
        b = np.ones(depth_filter_length) / depth_filter_length

        if a.size >= b.size:
            return np.correlate(a, b, mode="same")
        else:
            pad_width = int(np.ceil((b.size - a.size) / 2))
            a = np.pad(a, pad_width, "constant")
            return np.correlate(a, b, mode="same")[pad_width:-pad_width]

    def process(self, result: a121.Result) -> ProcessorResult:
        frame = result.subframes[self.subsweep_index]

        # Noise estimation

        nd = self.noise_est_diff_order

        noise_diff = np.diff(frame, n=nd, axis=0)
        noise = self._abs_dev(noise_diff, axis=0, subtract_mean=False)
        noise /= self.noise_norm_factor
        sf = self._dynamic_sf(self.noise_sf, self.update_index)
        self.lp_noise = sf * self.lp_noise + (1.0 - sf) * noise

        # Intra-frame part

        sweep_dev = self._abs_dev(frame, axis=0, ddof=1)

        sf = self._dynamic_sf(self.intra_sf, self.update_index)
        self.lp_intra_dev = sf * self.lp_intra_dev + (1.0 - sf) * sweep_dev

        norm_lp_intra_dev = np.divide(
            self.lp_intra_dev,
            self.lp_noise,
            out=np.zeros(self.num_distances),
            where=(self.lp_noise > 1.0),
        )

        intra = self._depth_filter(norm_lp_intra_dev, self.depth_filter_length)

        # Inter-frame part

        mean_sweep = np.abs(frame.mean(axis=0))

        sf = self._dynamic_sf(self.fast_sf, self.update_index)
        self.fast_lp_mean_sweep = sf * self.fast_lp_mean_sweep + (1.0 - sf) * mean_sweep

        sf = self._dynamic_sf(self.slow_sf, self.update_index)
        self.slow_lp_mean_sweep = sf * self.slow_lp_mean_sweep + (1.0 - sf) * mean_sweep

        inter_dev = np.abs(self.fast_lp_mean_sweep - self.slow_lp_mean_sweep)
        sf = self._dynamic_sf(self.inter_dev_sf, self.update_index)
        self.lp_inter_dev = sf * self.lp_inter_dev + (1.0 - sf) * inter_dev

        norm_lp_dev = np.divide(
            self.lp_inter_dev,
            self.lp_noise,
            out=np.zeros_like(self.lp_inter_dev),
            where=(self.lp_noise > 1.0),
        )

        norm_lp_dev *= np.sqrt(self.sweeps_per_frame)

        inter = self._depth_filter(norm_lp_dev, self.depth_filter_length)

        # Detector output

        depthwise_presence = self.inter_weight * inter + self.intra_weight * intra

        max_depthwise_presence = np.max(depthwise_presence)

        sf = self._dynamic_sf(self.output_sf, self.update_index)
        self.presence_score = sf * self.presence_score + (1.0 - sf) * max_depthwise_presence

        presence_detected = self.presence_score > self.threshold

        # TODO: self.presence_history will be removed in the future
        self.presence_history = np.roll(self.presence_history, -1)
        self.presence_history[-1] = self.presence_score

        if max_depthwise_presence > self.threshold:
            self.presence_distance_index = int(np.argmax(depthwise_presence))
            self.presence_distance = self.distances[self.presence_distance_index]

        self.update_index += 1

        extra_result = ProcessorExtraResult(
            frame=frame,
            mean_sweep=mean_sweep,
            fast_lp_mean_sweep=self.fast_lp_mean_sweep,
            slow_lp_mean_sweep=self.slow_lp_mean_sweep,
            lp_noise=self.lp_noise,
            inter=inter * self.inter_weight,
            intra=intra * self.intra_weight,
            depthwise_presence=depthwise_presence,
            presence_distance_index=self.presence_distance_index,
            presence_history=self.presence_history,
        )

        return ProcessorResult(
            presence_score=self.presence_score,
            presence_detected=presence_detected,
            presence_distance=self.presence_distance,
            extra_result=extra_result,
        )
